import numpy as np
from scipy.signal import welch as welch_sp
from scipy.signal import hilbert as hilbert_sp
from medusa.utils import check_dimensions


def hilbert(signal):
    """This method implements the Hilbert transform.

    Parameters
    ----------
    signal :  numpy.ndarray
        Input signal with shape [n_epochs x n_samples x n_channels].

    Returns
    -------
    hilb : numpy 3D matrix
        Analytic signal of x [n_epochs, n_samples, n_channels].
    """
    # Check dimensions
    signal = np.asarray(signal)
    signal = check_dimensions(signal)
    # Check errors
    if np.iscomplexobj(signal):
        raise ValueError("Signal must be real.")

    # Old tensorflow implementation
    # if tensorflow_integration.check_tf_config(autoconfig=True) and flag:
    #
    #     # Run the fft on the columns, not the rows.
    #     signal = tf.convert_to_tensor(signal, dtype=tf.complex128)
    #     signal = tf.transpose(tf.signal.fft(tf.transpose(signal)))
    #
    #     # Coeficients
    #     h = np.zeros(n)
    #     if (n > 0) and (2*np.fix(n/2) == n):
    #         # Even and nonempty
    #         h[0:int(n/2+1)] = 1
    #         h[1:int(n/2)] *= 2
    #     elif n > 0:
    #         # Odd and nonempty
    #         h[0] = 1
    #         h[1:int((n+1)/2)] = 2
    #
    #     tf_h = tf.constant(h, name='h', dtype=tf.float64)
    #     if len(signal.shape) == 2:
    #         reps = tf.Tensor.get_shape(signal).as_list()[-1]
    #         hs = tf.stack([tf_h]*reps, -1)
    #     elif len(signal.shape) == 1:
    #         hs = tf_h
    #     else:
    #         raise NotImplementedError
    #
    #     xc = signal * tf.complex(hs, tf.zeros_like(hs))
    #     return tf.transpose(tf.signal.ifft(tf.transpose(xc)))

    return hilbert_sp(signal, axis=1)


def power_spectral_density(signal, fs, segment_pct=80, overlap_pct=50,
                           window='boxcar'):
    """This method allows to compute the one-sided power spectral density (PSD)
    by means of Welch's periodogram method. This method wraps around
    scipy.signal.welch method to compute the PSD, allowing to pass epoched
    signals and defining the segment length and overlap in percentage,
    simplifying the use for specific purposes. For more advanced configurations,
    use the original scipy (or equivalent) function.

    Parameters
    ----------
    signal : numpy nd array
        Signal with shape [n_epochs x n_samples x n_channels].
    fs : int
        Sampling frequency of the signal
    segment_pct: float
        Percentage of the signal (n_samples) used to calculate the FFT. Default:
        80% of the signal.
    overlap_pct: float
        Percentage of overlap (n_samples) for the Welch method. Default: 50% of
        the signal.
    window:
        Desired window to use. See scipy.signal.welch docs for more details

    Returns
    -------
    f : numpy 1D array
        Array of sampled frequencies.
    psd: numpy 2D array
        PSD of the signal with shape [n_epochs, n_samples, n_channels]
    """

    # Check signal dimensions
    signal = check_dimensions(signal)
    # Get the number of samples for the PSD length
    n_samp = signal.shape[1]
    # Get nperseg and noverlap
    nperseg = n_samp * segment_pct / 100
    noverlap = n_samp * overlap_pct / 100
    # Compute the PSD
    f, psd = welch_sp(signal, fs=fs, window=window, nperseg=nperseg,
                      noverlap=noverlap, axis=1)
    return f, psd


def normalize_psd(psd, norm='rel'):
    """Normalizes the PSD using different methods.

    Parameters
    ----------
    psd: numpy array or list
        Power Spectral Density (PSD) of the signal with shape [samples],
        [samples x channels] or [epochs x samples x channels]. It assumes PSD
        is the one-sided spectrum.
    norm: string
        Normalization to be performed. Choose z for z-score or rel for
        relative power.

    """
    # To numpy arrays
    psd = np.array(psd)

    # Check errors
    if len(psd.shape) != 3:
        raise Exception('Parameter psd must have shape [n_epochs x n_samples x '
                        'n_channels]')

    if norm == 'rel':
        p = np.sum(psd, axis=1, keepdims=True)
        psd_norm = psd / p
    elif norm == 'z':
        m = np.mean(psd, keepdims=True, axis=1)
        s = np.std(psd, keepdims=True, axis=1)
        psd_norm = (psd - m) / s
    else:
        raise Exception('Unknown normalization. Choose z or rel')

    return psd_norm





"""
In this module you will find useful functions and classes to apply on-line
Neurofeedback models. Each model is based on different features to be used
as target to train. Enjoy!

@author: Diego Marcos-Martínez
"""

# Built-in imports
from abc import ABC, abstractmethod
import concurrent

# External imports
import numpy as np
import scipy.signal

# Medusa imports
import medusa as mds
from medusa import components
from medusa import meeg
from medusa.spatial_filtering import LaplacianFilter, car
from medusa.connectivity.phase_connectivity import phase_connectivity
from medusa.connectivity.amplitude_connectivity import __aec_ort_cpu as aec
from medusa.graph_theory import degree
from medusa.artifact_removal import reject_noisy_epochs
from medusa.epoching import get_epochs_of_events
from medusa.local_activation.spectral_parameteres import absolute_band_power


class SignalPreprocessing(components.ProcessingMethod):
    """
    Common preprocessing applied in Neurofeedback applications.
    It is composed by a frequency IIR filter followed by a spatial
    filters. Functions are adapted to filter the signal in more than one frequency
    range, if necessary.
    """

    def __init__(self, filter_dict=None, montage=None, target_channels=None,
                 laplacian=False, car=False, n_cha_lp=None):
        super().__init__(prep_fit_transform=['signal'],
                         prep_transform=['signal'],
                         narrow_transform=['signal'])

        # Error check
        if not filter_dict:
            raise ValueError('[SignalPreprocessing] Filter dict parameter '
                             '"filter_dict" must be a list containing all '
                             'necessary information to perform the filtering!. '
                             'The information should be: type and cutoff')
        for filter in filter_dict:
            if not isinstance(filter, dict):
                raise ValueError('[SignalPreprocessing] Each filter must '
                                 'be a dict()!')
            if 'cutoff' not in filter or 'type' not in filter:
                raise ValueError('[SignalPreprocessing] Each filter must '
                                 'be a dict() containing the following keys: '
                                 '"cutoff"and "type"!')
            if filter['type'] != 'training' and  filter['type'] != 'artifact':
                raise ValueError('[SignalPreprocessing] "type" must be "training"'
                                 'or "artifact".')
        if not montage:
            raise ValueError('[SignalPreprocessing] Pre-processing parameter'
                             '"montage" must be a dict containing all'
                             'labels of channels and montage standard key')
        if laplacian and target_channels is None:
            raise ValueError('[SignalPreprocessing] Laplacian filter needs to '
                             'define "target_channels" parameter.')

        # Parameters
        self.filter_dict = filter_dict
        self.l_cha = montage.l_cha
        self.target_channels = target_channels
        self.montage = montage
        self.perform_car = car
        self.perform_laplacian = laplacian

        # Variables
        self.artifact_iir_filters = []
        self.target_iir_filters = []
        self.offset_line_removal = None
        self.laplacian_filter = None

    def fit(self, fs):
        """
        Fits the IIR filter and Laplacian spatial filter (if selected) for signal
        preprocessing stage.

        Parameters
        ----------
        fs: float
            Sampling rate in Hz.
        """

        # Fit Spectral Filters (Predefined to be optimal)
        self.offset_line_removal = mds.IIRFilter(order=3,
                                                 cutoff=[0.5, 40],
                                                 btype='bandpass',
                                                 filt_method='sosfiltfilt')
        self.offset_line_removal.fit(fs, len(self.l_cha))

        # Define filters for filtering over epochs
        for filter in self.filter_dict:
            for f in filter['cutoff']:
                iir = mds.IIRFilter(order=1,
                                    cutoff=f,
                                    btype='bandpass',
                                    filt_method='sosfiltfilt')

                if self.target_channels is None:
                    iir.fit(fs, len(self.l_cha))
                else:
                    iir.fit(fs, len(self.target_channels))
                if filter['type'] == 'artifact':
                    self.artifact_iir_filters.append(iir)
                elif filter['type'] == 'training':
                    self.target_iir_filters.append(iir)

        # Fit Laplacian Filter
        if self.perform_laplacian:
            if len(self.montage.l_cha) >= 5:
                self.laplacian_filter = LaplacianFilter(self.montage, mode='auto')
                self.laplacian_filter.fit_lp(l_cha_to_filter=self.target_channels)

    def prep_transform(self, signal, parallel_computing=True):
        """
        Transforms an EEG signal applying IIR filter. It also applies CAR and
        Laplacian spatial filter sequentially if desired.

        Parameters
        ----------
        signal: list or numpy.ndarray
            Signal to transform. Shape [n_samples x n_channels]
        parallel_computing: bool
            If true, it filters the signal concurrently

        Returns
        -------
        signal_: numpy.ndarray
            Original signal with power line and offset removed, and spatially
            filtered if chosen. [n_samples, n_channels].
        signal_artifacts: numpy.ndarray
            If it has been chosen to reject artifact sections, this matrix
            contains the filtered signal in the frequency bands associated
            with these artifacts. [n_artifact_bands, n_samples, n_channels].
        """
        # Initialize variable
        n_samples = signal.shape[0]

        if len(self.artifact_iir_filters) == 0:
            signal_artifacts = None
        else:
            if self.target_channels is None:
                signal_artifacts = np.empty(
                    (
                    len(self.artifact_iir_filters), n_samples, len(self.l_cha)))
            else:
                signal_artifacts = np.empty(
                    (len(self.artifact_iir_filters), n_samples,
                     len(self.target_channels)))

        signal_ = self.offset_line_removal.transform(signal)

        # Spatial filtering
        if self.perform_car:
            signal_ = car(signal_)
        if self.perform_laplacian:
            # Check if surface laplacian filter cannot be performed
            if self.laplacian_filter is not None:
                signal_ = self.laplacian_filter.apply_lp(signal_)
            else:
                signal_ = signal_[:,self.montage.get_cha_idx_from_labels(self.target_channels)]

        signal__ = signal_.copy()

        if signal_artifacts is not None:
            # Filter only the target channels in Power Based models
            if signal__.shape[1] != signal_artifacts.shape[2]:
                signal__ = signal__[:,self.montage.get_cha_idx_from_labels(
                    self.target_channels)]

            # Frequency filtering on artifact-related bands
            if parallel_computing:
                filt_threads = []
                for filter in self.artifact_iir_filters:
                    t = components.ThreadWithReturnValue(target=
                                                         filter.transform,
                                                         args=(signal__,))
                    filt_threads.append(t)
                    t.start()

                for filt_idx, thread in enumerate(filt_threads):
                    signal_artifacts[filt_idx, :, :] = thread.join()
            else:
                for filt_idx, filter in enumerate(self.artifact_iir_filters):
                    signal_artifacts[filt_idx, :, :] = filter.transform(
                        signal__[np.newaxis, :, :])
        return signal_, signal_artifacts

    def prep_fit_transform(self, fs, signal):
        """
        Fits the IIR filter and transforms an EEG signal applying IIR
        filter and spatial filters sequentially.

        Parameters
        ----------
        fs: float
            Sampling rate in Hz.
        n_cha_lp: int
            Number of nearest channels to compute Laplacian spatial
            filter (Auto mode).
        signal: list or numpy.ndarray
            Signal to transform. Shape [n_samples x n_channels]

        Returns
        -------
        signal: numpy.ndarray
            Original signal with power line and offset removed, and spatially
            filtered if chosen. [n_samples, n_channels].
        signal_artifacts: numpy.ndarray
            If it has been chosen to reject artifact sections, this matrix
            contains the filtered signal in the frequency bands associated
            with these artifacts. [n_artifact_bands, n_samples, n_channels].
        """
        self.fit(fs)
        return self.prep_transform(signal)

    def narrow_transform(self, signal, parallel_computing=True):
        """
        Applies the IIR filter for narrow band filtering.

        Parameters
        ----------
        signal: list or numpy.ndarray
            Signal to transform. Shape [n_samples x n_channels].
        parallel_computing: bool
            If true, it filters the signal concurrently.

        Returns
        -------
        signal: numpy.ndarray
            Signal filtered in the training band [n_samples, n_channels].
        """
        f_signals = np.empty(
            (len(self.target_iir_filters), signal.shape[0], signal.shape[1]))
        if parallel_computing:
            filt_threads = []
            for filter in self.target_iir_filters:
                t = components.ThreadWithReturnValue(target=
                                                     filter.transform,
                                                     args=(signal,))
                filt_threads.append(t)
                t.start()

            for filt_idx, thread in enumerate(filt_threads):
                f_signals[filt_idx, :, :] = thread.join()
        else:
            for filt_idx, filter in enumerate(self.target_iir_filters):
                f_signals[filt_idx, :, :] = filter.transform(signal)

        return np.squeeze(f_signals)

def ignore_noisy_windows(signals, thresholds, pct_tol):
    """
    This function check if a specific signal segment contains noise above the
    pre-established thresholds.
    Parameters
    ----------
    signals: numpy.ndarray
        Array containing the signal filtered in the frequency bands associated to
        artifacts to avoid. [n_artifact_bands, n_samples, n_channels].
    thresholds: numpy.ndarray
        Array containing the variance thresholds related to artifacts to avoid.
    pct_tol: numpy.ndarray
        Array containing variance increase (in percentage) tolerated.

    Returns
    -------

    """
    # Check if power in forbidden bands is over thresholds
    over_var = np.sum(np.std(signals, axis=1).mean(axis=-1) >=
                      (1 + pct_tol) * thresholds)
    if over_var < 1:
        return True
    else:
        return False


def make_windows(signal, fs, update_feature_window, update_rate,
                 n_cha=1, n_samp=2, k=4, reject=True):
    """

    Parameters
    ----------
    signal: numpy.ndarray
        Signal to be converted into epochs. [n_samples, n_channels].
    fs: int or float
        Sampling rate.
    update_feature_window: int or float
        Time window taken for the calculation of the characteristic (in seconds).
    update_rate: int or float
        Feedback update time in online mode.
    n_cha: int
        Threshold number of channels meeting the rejection condition to reject
        the epoch.
    n_samp: int
        Threshold number of samples meeting the rejection condition to reject
        the epoch.
    k: int
        Standard deviation of the signal. Used in the definition of the
        rejection criterion.
    reject: bool
        If true, it returns the epochs that have not been reject. Else, it
        returns the whole windowed signal.

    Returns
    -------
    good_epochs: numpy.ndarray
        Array containing the signal divided into epochs that are not noisy.
        [n_epochs, n_samples, n_channels].
    ind: numpy.ndarray
        Array containing bools. True for epochs that were rejected and False for
         epochs that were not.

    """
    if len(signal.shape) == 1:
        signal = signal[:, np.newaxis]
    # Define necessary parameters
    s_duration = signal.shape[0] / fs
    s_mean = np.mean(signal, axis=0)
    s_std = np.std(signal, axis=0)

    # Set onsets vector
    onsets = np.arange(0, s_duration - update_feature_window, update_rate)
    s_windowed = get_epochs_of_events(np.arange(0, s_duration, 1 / fs), signal,
                                      onsets, fs,
                                      [0, update_feature_window * 1000])
    # Return windows without discarding
    if not reject:
        return s_windowed
    pct_rejected, good_epochs, idx = reject_noisy_epochs(s_windowed, s_mean,
                                                         s_std, k, n_samp,
                                                         n_cha)
    return good_epochs, idx


class ConnectivityExtraction(components.ProcessingMethod):
    """
    Functional Connectivity-based features to extract from user's EEG.
    """

    def __init__(self, l_baseline_t=5, fs=250, update_feature_window=2,
                 update_rate=0.25, fc_measure=None, mode=None, montage=None,
                 target_channels=None, pct_tol=0.9):
        """
        Class constructor

        l_baseline_t: int
            Time employed to calculate the number of samples to obtain baseline
            connectivity parameter. In seconds.
        fs: int or float
            Sample rate of the recording.
        update_feature_window: int or float
            Length in seconds of the temporal window applied to calculate
            the feature.
        update_rate: int or float
            Feedback update time in online mode.
        fc_measure: str
            "WPLI" or "AECORT". Measure of Functional Connectivity to calculate.
        mode: str
            "Global coupling", "Strength" or "Coupling". Information extracted
            from adjacency matrix.
        montage: EEGChannelSet
        target_channels: list or None
            List containing the labels of the target channels.
        pct_tol: numpy.ndarray
        Array containing variance increase (in percentage) tolerated.
        """

        super().__init__(ext_feature=['conn_value'])

        # Check errors
        if not montage:
            raise ValueError('[ConnectivityExtraction] "montage parameter"'
                             ' must be a dict containing all'
                             'labels of channels and montage standard key')
        if fc_measure != "WPLI" and fc_measure != "AECORT":
            raise ValueError('[ConnectivityExtraction] Invalid functional '
                             'connectivity measure. Available measures are '
                             '"WPLI" and "AECORT".')
        if mode != "Global coupling" and mode != "Strength" and mode != \
                "Coupling":
            raise ValueError('[ConnectivityExtraction] Invalid mode. '
                             'Available modes are '
                             '"Global coupling", "Strength" and "Coupling".')
        if target_channels is None:
            if mode == "Strength":
                raise UserWarning(
                    '[ConnectivityExtraction] Using "Strength" mode'
                    'without defining target channels. Average strength'
                    'of all channels will be returned instead.')
            if mode == "Coupling":
                mode = "Global coupling"
                raise UserWarning(
                    '[ConnectivityExtraction] Using "Coupling" mode'
                    'without defining target channels. Global coupling of all '
                    'channels will be returned instead.')

            self.target_channels = target_channels
        else:
            self.target_channels = montage.get_cha_idx_from_labels(
                target_channels)

        self.fc_measure = fc_measure
        self.mode = mode
        self.fs = fs
        self.l_baseline_t = l_baseline_t
        self.montage = montage
        self.update_feature_window = update_feature_window
        self.update_rate = update_rate
        self.w_signal_samples = int(update_feature_window * self.fs)
        self.w_signal_samples_calibration = int((self.l_baseline_t)* self.fs)
        self.pct_tol = pct_tol
        self.thresholds = None
        self.baseline_value = None

    def set_baseline(self, signal, signal_artifacts, filtered_signal):
        """
        This functions establish the baseline value.
        Parameters
        ----------
        signal: numpy.ndarray
            Original signal pre-processed (offset and power-line removed, and CAR)
            [n_samples, n_channels].
        signal_artifacts: numpy.ndarray
            Array containing the signal filtered in each frequency band associated
            to the artifacts  to avoid. [n_artifact_bands, n_samples, n_channels].
        filtered_signal: numpy.ndarray
            Signal filtered in the narrow band for training. [n_samples, n_channels].
        Returns
        -------
        baseline_value: float
        """
        # Delete borders to avoid effect borders
        dif_window = round((signal.shape[0] - self.w_signal_samples_calibration)/2)
        if dif_window <= 0:
            raise ValueError('The duration of the signal for the baseline '
                             'calculation should be slightly longer than the '
                             'window time taken for the baseline calculation. ')
        epochs_original, index = make_windows(signal[dif_window:-dif_window, :]
                                , self.fs, self.update_feature_window,
                                self.update_rate)
        filtered_epochs = make_windows(
                filtered_signal[dif_window:-dif_window, :], self.fs,
                self.update_feature_window, self.update_rate,
                reject=False)[~index, :, :]

        adj_mat = self.calculate_adj_mat(filtered_epochs)

        # Parallel computing baseline values
        filt_threads = []
        baseline_values = []
        for epoch_mat in adj_mat:
            t = components.ThreadWithReturnValue(target=self.calculate_feature,
                                                 args=(epoch_mat,))
            filt_threads.append(t)
            t.start()

        for filt_idx, thread in enumerate(filt_threads):
            baseline_values.append(thread.join())

        self.baseline_value = np.mean(baseline_values)

        # Define artifact related thresholds
        if signal_artifacts is not None:
            self.thresholds = np.std(
                signal_artifacts[:, -self.w_signal_samples_calibration:,
                :], axis=1).mean(axis=-1)
        return self.baseline_value

    def ext_feature(self, signal, signal_artifacts):
        """
        Function for extracting FC values in online mode.
        Parameters
        ----------
        signal: numpy.ndarray
            Signal filtered in the narrow band for training. [n_samples, n_channels].
        signal_artifacts: numpy.ndarray or None
            Array containing the signal filtered in each frequency band associated
            to the artifacts  to avoid. [n_artifact_bands, n_samples, n_channels].
        Returns
        -------
        c_value: float
        """
        if self.baseline_value is None:
            raise ValueError(
                '[ConnectivityExtraction] Calibration not performed.')
        # Delete borders
        dif_window = round((signal.shape[0] - self.w_signal_samples)/2)
        if dif_window <= 0:
            raise ValueError('The duration of the signal for feedback '
                             'calculation should be slightly longer than the '
                             'window time taken for the feedback calculation. ')

        adj_mat = self.calculate_adj_mat(signal[dif_window:-dif_window, :])
        c_value = self.calculate_feature(np.squeeze(adj_mat)) \
                  - self.baseline_value
        # Check if artifact bands are defined
        if signal_artifacts is not None:
            if ignore_noisy_windows(signal_artifacts[:,dif_window:-dif_window,
                                    self.target_channels], self.thresholds,
                                    self.pct_tol):
                return c_value
            else:
                return None
        return c_value

    def calculate_adj_mat(self, signal):
        """
        This function calculates the adjacency matrix depending on the FC mode.
        Parameters
        ----------
        signal: numpy.ndarray
            Signal filtered in the narrow band for training.
            [n_epochs, n_samples, n_channels] or [n_samples, n_channels].
        Returns
        -------
        adj_mat: numpy.ndarray
            [n_epochs, n_channels, n_channels].
        """
        # Calculate adjacency matrix depending on FC measure chosen
        adj_mat = None
        if self.fc_measure == "WPLI":
            adj_mat = phase_connectivity(signal,'wpli')
        # This is under development
        elif self.fc_measure == "AECORT":
            adj_mat = aec(signal)
        return adj_mat

    def calculate_feature(self, adj_mat):
        """
        Calculates Graph metric from adjacency matrix.

        Parameters
        ----------
        adj_mat: numpy.ndarray
            [n_channels, n_channels].
        """
        # Calculate the baseline value depending on mode chosen
        if self.mode == "Global coupling":
            tri_l_idx = np.tril_indices(adj_mat.shape[0], -1)
            return np.nanmean(np.asarray(adj_mat)[tri_l_idx])
        elif self.mode == "Strength":
            if self.target_channels is None:
                return np.mean(degree.degree(np.asarray(adj_mat),
                                             'CPU'))
            else:
                return np.mean(degree.degree(np.asarray(adj_mat),
                                             'CPU')[
                                   self.target_channels])
        elif self.mode == "Coupling":
            return self.mean_coupling(adj_mat)

    def mean_coupling(self, adj_mat):
        """
        This function calculates the connectivity values between all the
           target channels and average it value.

        Parameters
        ----------
        adj_mat: numpy.ndarray
            [n_channels, n_channels].
        """
        c = []
        for ind, ch_ind_1 in enumerate(self.target_channels[:-1]):
            for ch_ind_2 in self.target_channels[ind + 1:]:
                c.append(np.array(adj_mat[ch_ind_1, ch_ind_2]))
        return np.mean(c)


class PowerExtraction(components.ProcessingMethod):
    """
    Power-based features to extract from user's EEG.
    """

    def __init__(self, l_baseline_t=5, fs=250, update_feature_window=2,
                 update_rate=0.25, pct_tol = 0.9, f_dict=None, mode=None):
        """
        Class constructor

        l_baseline_t: int
            Time employed to calculate the number of samples to obtain baseline
            power parameter. In seconds.
        fs: int or float
            Sample rate of the recording.
        update_feature_window = int
            Length in seconds of the temporal window applied to calculate
            the feature.
        update_rate: int or float
            Feedback update time in online mode.
        f_dict: dict
            Dict containing the frequency bands associated to training band
            and artifacts to avoid.
        pct_tol: numpy.ndarray
            Array containing variance increase (in percentage) tolerated.
        mode: str
            "single" or "ratio"
        """

        super().__init__(band_power=['band power'])

        self.mode = mode
        self.fs = fs
        self.l_baseline_t = l_baseline_t
        self.update_feature_window = update_feature_window
        self.update_rate = update_rate
        self.w_signal_samples = int(update_feature_window * self.fs)
        self.w_signal_samples_calibration = int(self.l_baseline_t * self.fs)
        self.f_dict = f_dict
        self.pct_tol = pct_tol
        self.baseline_power = None
        self.thresholds = None

    def set_baseline(self, signal, signal_artifacts):
        """
        This function sets the power baseline, given the already filtered EEG
        containing the calibration phase. Also, takes into account the
        Neurofeedback training mode, so performs different baseline calculations
        depending on the mode set.

        Parameters
        ----------
        signal: numpy.ndarray
            EEG already pre-processed. [n_samples, n_channels].
        signal_artifacts: numpy.ndarray
            Signal filtered in the frequency bands associated to the artifacts
            to be avoided. [n_artifact_bands, n_samples, n_channels].
        Returns
        ------
        baseline_power: float
        """
        # Delete borders
        dif_window = round(
            (signal.shape[0] - self.w_signal_samples_calibration) / 2)
        if dif_window <= 0:
            raise ValueError('The duration of the signal for the baseline '
                             'calculation should be slightly longer than the '
                             'window time taken for the baseline calculation.')
        epochs, _ = make_windows(signal[dif_window:-dif_window, :]
                                 , self.fs, self.update_feature_window,
                                 self.update_rate)
        _, psd = scipy.signal.welch(epochs, self.fs, 'hamming',
                                    self.w_signal_samples,
                                    axis=1,scaling='density')

        b_power = self.power(psd)

        # Define artifact related thresholds
        if signal_artifacts is not None:
            self.thresholds = np.std(
                signal_artifacts[:, -self.w_signal_samples_calibration:,
                :], axis=1).mean(axis=-1)

        if self.mode == 'single':
            self.baseline_power = b_power[0]

        elif self.mode == 'ratio':
            self.baseline_power = b_power[0] / b_power[1]

        return self.baseline_power

    def band_power(self, signal, signal_artifacts):
        """
        This function returns the band power from Power Spectral Density.
        If signal noise is above the pre-established thresholds, this function
        will return None.

        Parameters
        ----------
        signal: numpy.ndarray
            Signal pre-processed. [n_samples, n_channels].
        signal_artifacts: numpy.ndarray
            Signal filtered in the frequency bands associated to the artifacts
            to be avoided. [n_artifact_bands, n_samples, n_channels].
        Returns
        ------
        b_power: float or None
        """

        if self.baseline_power is None:
            raise ValueError('[PowerExtraction] Calibration not performed.')

        dif_window = round(
            (signal.shape[0] - self.w_signal_samples) / 2)
        if dif_window <= 0:
            raise ValueError('The duration of the signal for the feedback '
                             'calculation should be slightly longer than the '
                             'window time taken for the feedback calculation. ')

        _, psd = scipy.signal.welch(signal[dif_window:-dif_window], self.fs, 'hamming',
                                    self.w_signal_samples,
                                    axis=0,scaling='density')
        b_power_uncorrected = self.power(psd)
        if self.mode == 'single':
            b_power = b_power_uncorrected[0] - self.baseline_power

        elif self.mode == 'ratio':
            b_power = b_power_uncorrected[0] / b_power_uncorrected[
                1] - self.baseline_power

        # Check if artifact bands are defined
        if signal_artifacts is not None:
            if signal_artifacts is not None:
                if ignore_noisy_windows(signal_artifacts, self.thresholds,
                                        self.pct_tol):
                    return b_power
                else:
                    return None
        return b_power

    def power(self, psd):
        """
        This function calculates power from Power Spectral Density
        Parameters
        ----------
        psd: numpy.ndarray
            [n_epochs, n_samples, n_channels].

        Returns
        -------
        powers: numpy.ndarray
            [n_training_bands].
        """
        # Check if psd has epochs dimension
        if len(psd.shape) == 2:
            psd = psd[np.newaxis,:,:]
        bands = []
        # Extract training bands limits
        for dict in self.f_dict:
            if dict['type'] == 'training':
                bands.append(dict['cutoff'])
        powers = np.zeros(len(bands))

        # Calculate band power relative to the whole bandwidth
        for idx, band in enumerate(bands):
            for b in band:
                powers[idx] += np.mean(np.mean(absolute_band_power(psd, self.fs,
                                                                   b),
                                               axis=0))
        return powers


class ConnectivityBasedNFTModel(components.Algorithm):
    def __init__(self, fs, filter_dict, l_baseline_t, update_feature_window,
                 update_rate, montage, target_channels, fc_measure, mode,
                 apply_car, pct_tol_ocular=None,pct_tol_muscular=None):
        super().__init__(calibration=['baseline_value'],
                         training=['feedback_value'])
        """
        Pipeline for Connectivity-based Neurofeedback training. This class
        inherits from components.Algorithm. Therefore, it can be used to create
        standalone algorithms that can be used in compatible apps from
        medusa-platform for online experiments. See components.Algorithm to know
        more about this functionality.
        """

        # Settings
        self.fs = fs
        self.filter_dict = filter_dict
        self.l_baseline_t = l_baseline_t
        self.update_feature_window = update_feature_window
        self.montage = montage
        self.target_channels = target_channels
        self.fc_measure = fc_measure
        self.mode = mode
        self.apply_car = apply_car

        # Variables
        self.baseline_value = None
        self.pct_tol = None

        # Set percentage tolerance to noisy signal
        if pct_tol_ocular is None and pct_tol_muscular is not None:
            self.pct_tol = pct_tol_muscular
        elif pct_tol_ocular is not None and pct_tol_muscular is None:
            self.pct_tol = pct_tol_ocular
        elif pct_tol_ocular is not None and pct_tol_muscular is not None:
            self.pct_tol = np.array([pct_tol_ocular, pct_tol_muscular])

        # Check filter dict
        if not self.check_cutoff_settings():
            raise Exception('The number of frequency bands selected does not '
                            'match the Neurofeedback mode.')

        # Add Pre-processing and Feature Extraction methods
        self.add_method('prep_method',
                        SignalPreprocessing(filter_dict=self.filter_dict,
                                            montage=self.montage,
                                            target_channels=None,
                                            car=self.apply_car))
        self.add_method('feat_ext_method',
                        ConnectivityExtraction(fs=self.fs, l_baseline_t=self.l_baseline_t,
                                               update_feature_window=update_feature_window,
                                               fc_measure=self.fc_measure,
                                               mode=self.mode,montage=self.montage,
                                               target_channels=self.target_channels,
                                               pct_tol=self.pct_tol,update_rate=update_rate))

    def calibration(self, eeg):
        """
        It pre-process eeg, gets signal filtered in artifact-related bands and
        filters the pre-processed eeg in training band. Then, it calculates the
        baseline value.
        Parameters
        ----------
        eeg: numpy.ndarray
            [n_samples, n_channels]
        Returns
        -------
        baseline_value: float
        """
        original_signal, signal_artifacts = self.get_inst('prep_method').\
            prep_fit_transform(
            signal=eeg,
            fs=self.fs)
        narrow_filtered_signal = self.get_inst('prep_method').\
            narrow_transform(signal=original_signal)
        self.baseline_value = self.get_inst('feat_ext_method').\
            set_baseline(signal=original_signal,signal_artifacts=signal_artifacts,
                         filtered_signal=narrow_filtered_signal)

    def training(self, eeg):
        """
        It pre-process eeg, gets signal filtered in artifact-related bands and
        filters the pre-processed eeg in training band. Then, it calculates the
        feedback value.
        Parameters
        ----------
        eeg: numpy.ndarray
            [n_samples, n_channels]
        Returns
        -------
        feedback_value: float
        """
        original_signal, signal_artifacts = self.get_inst('prep_method').prep_transform(
            signal=eeg)
        narrow_filtered_signal = self.get_inst('prep_method'). \
            narrow_transform(signal=original_signal)
        feedback_value = self.get_inst('feat_ext_method').ext_feature(
            signal=narrow_filtered_signal,signal_artifacts=signal_artifacts)
        return feedback_value

    def check_cutoff_settings(self):
        """
        Function to check the correct definition of training band dictionary.
        """
        target_bands = 0
        for filter in self.filter_dict:
            if filter['type'] == 'training':
                target_bands += 1
        if target_bands == 1:
            return True
        else:
            return False


class PowerBasedNFTModel(components.Algorithm):
    def __init__(self, fs, filter_dict, l_baseline_t, update_feature_window,
                 update_rate,montage, target_channels, mode, apply_car,
                 apply_laplacian, pct_tol_ocular=None, pct_tol_muscular=None):
        """
        Pipeline for Power-based Neurofeedback training. This class
        inherits from components.Algorithm. Therefore, it can be used to create
        standalone algorithms that can be used in compatible apps from
        medusa-platform for online experiments. See components.Algorithm to know
        more about this functionality.
        """
        super().__init__(calibration=['baseline_parameters'],
                         training=['feedback_value'])

        """
        Class constructor
        """

        # Settings
        self.fs = fs
        self.filter_dict = filter_dict
        self.l_baseline_t = l_baseline_t
        self.update_feature_window = update_feature_window
        self.montage = montage
        self.target_channels = target_channels
        self.mode = mode
        self.apply_car = apply_car
        self.apply_laplacian = apply_laplacian

        # Init variables
        self.baseline_value = None
        self.pct_tol = None

        # Set percentage tolerance to noisy signal
        if pct_tol_ocular is None and pct_tol_muscular is not None:
            self.pct_tol = pct_tol_muscular
        elif pct_tol_ocular is not None and pct_tol_muscular is None:
            self.pct_tol = pct_tol_ocular
        elif pct_tol_ocular is not None and pct_tol_muscular is not None:
            self.pct_tol = np.array([pct_tol_ocular, pct_tol_muscular])

        # # Check correct filter dict definition
        if not self.check_cutoff_settings():
            raise Exception('The number of frequency bands selected does not '
                            'match the Neurofeedback mode.')

        self.add_method('prep_method',
                        SignalPreprocessing(filter_dict=self.filter_dict,
                                            montage=self.montage,
                                            target_channels=self.target_channels,
                                            laplacian=self.apply_laplacian,
                                            car=self.apply_car))
        self.add_method('feat_ext_method',
                        PowerExtraction(fs=fs, l_baseline_t=l_baseline_t,
                                        update_feature_window=update_feature_window,
                                        update_rate=update_rate,mode=self.mode,
                                        pct_tol=self.pct_tol,f_dict=filter_dict))

    def calibration(self, eeg, **kwargs):
        """
        It pre-process eeg and gets signal filtered in artifact-related bands.
        Then, it calculates the baseline value.
        Parameters
        ----------
        eeg: numpy.ndarray
            [n_samples, n_channels]
        Returns
        -------
        baseline_value: float
        """
        original_signal, signal_artifacts = self.get_inst('prep_method').\
            prep_fit_transform(signal=eeg, fs=self.fs)
        self.baseline_value = self.get_inst('feat_ext_method').set_baseline(
            signal=original_signal,signal_artifacts=signal_artifacts)

    def training(self, eeg):
        """
        It pre-process eeg, gets signal filtered in artifact-related bands. Then,
        it calculates the feedback value.
        Parameters
        ----------
        eeg: numpy.ndarray
            [n_samples, n_channels]
        Returns
        -------
        feedback_value: float
        """
        original_signal, signal_artifacts = self.get_inst('prep_method').\
            prep_transform(signal=eeg)
        feedback_value = self.get_inst('feat_ext_method').band_power(
            signal=original_signal, signal_artifacts=signal_artifacts)
        return feedback_value

    def check_cutoff_settings(self):
        """
        Function to check the correct definition of training band dictionary.
        """
        target_bands = 0
        for filter in self.filter_dict:
            if filter['type'] == 'training':
                target_bands += 1
        if self.mode == 'single':
            if target_bands == 1:
                return True
            else:
                return False
        elif self.mode == 'ratio':
            if target_bands == 2:
                return True
            else:
                return False

class NeurofeedbackData(components.ExperimentData):
    """Experiment info class for Neurofeedback training experiments. It records
    the important events that take place during a Neurofeedback run,
    allowing offline analysis."""

    def __init__(self, run_onsets, run_durations, run_success, run_pauses,
                 run_restarts, medusa_nft_app_settings, nft_values, nft_times,
                 nft_baseline):

        self.run_onsets = run_onsets
        self.run_durations = run_durations
        self.run_success = run_success
        self.run_pauses = run_pauses
        self.run_restarts = run_restarts
        self.medusa_nft_app_settings = medusa_nft_app_settings
        self.nft_values = nft_values
        self.nft_times = nft_times
        self.nft_baseline = nft_baseline

    def to_serializable_obj(self):
        rec_dict = self.__dict__
        for key in rec_dict.keys():
            if type(rec_dict[key]) == np.ndarray:
                rec_dict[key] = rec_dict[key].tolist()
        return rec_dict

    @classmethod
    def from_serializable_obj(cls, dict_data):
        return cls(**dict_data)



from medusa import transforms
from medusa import signal_generators


def test_power_spectral_density():
    generator = signal_generators.SinusoidalSignalGenerator(
        fs=200, freqs=[1, 5], noise_type=None)
    chunk = generator.get_chunk(10, 5)
    f, psd = transforms.power_spectral_density(
        signal=chunk, fs=200, segment_pct=80, overlap_pct=50, window='boxcar')
    assert psd.shape == (1, 801, 5)
    assert psd[0, 8, 1] == 4


if __name__ == '__main__':
    test_power_spectral_density()




from sklearn.preprocessing import OneHotEncoder
import numpy as np


def categorical_labels(one_hot_labels):
    cat_labels = np.argmax(one_hot_labels, axis=1)
    return cat_labels


def one_hot_labels(categorical_labels):
    enc = OneHotEncoder(handle_unknown='ignore')
    one_hot_labels = \
        enc.fit_transform(categorical_labels.reshape(-1, 1)).toarray()

    return one_hot_labels


def k_fold_split(x, y, k, keys=None, shuffle=False):
    """
    Special implementation of k fold splitting that allows to split the dataset
    into k folds for cross validation in function of keys array.

    It returns a list with the dataset for each iteration (k iterations).

    Parameters
    ----------
    x: numpy array or list
        Training set data. Axis 0 represents each observation. Features could
        have one or more dimensions. For instance, [observations x eeg samples],
        [observations x eeg samples x channels]
    y: numpy array or list
        Training set labels.
    k: int
        Number of folds to split the dataset
    keys: numpy array or list
        Keys to split the dataset. If None, the dataset is splitted considering
        each observation independently. If not None, each position of keys
        array identifies the set that owns the observation. For instance, This
        is useful to split the dataset by subjects or trials.
    shuffle: boolean
        True if you want to shuffle the dataset randomly.

    Returns
    -------
    sets: list
        List that contains a dict with the train and test set for each iteration
        of the k-fold algorithm.

    Examples
    --------
    >>> k_fold_iter = k_fold_split(x, y, k)
    >>> k_fold_acc = 0
    >>> for iter in k_fold_iter:
    >>>     model.fit(iter[x_train, y_train])
    >>>     y_test_pred = model.predict(iter[x_test, y_test])
    >>>     k_fold_acc += np.sum(y_test_pred == y_test)/len(y_test)
    >>>     k_fold_acc = k_fold_acc/len(k_fold_iter)
    """
    # Convert to numpy arrays
    x = np.array(x)
    y = np.array(y)
    # If keys is None, each observation is treated independently
    if keys is None:
        keys = np.arange(len(x))
    else:
        keys = np.array(keys)
    if keys.shape[0] != x.shape[0] or keys.shape[0] != y.shape[0]:
        raise ValueError("Dimensions of x, y and keys arrays must match along"
                         " axis 0.")
    # Divide keys array in k folds
    keys_values = np.unique(keys)
    if shuffle:
        np.random.shuffle(keys_values)
    keys_folds = np.array_split(keys_values, k)
    # Divide the dataset
    k_fold_iter = list()
    for i in range(k):
        idx = np.isin(keys, keys_folds[i]).nonzero()
        # Get train set
        x_train = np.delete(x, idx, axis=0)
        y_train = np.delete(y, idx, axis=0)
        # Get test set
        x_test = x[idx]
        y_test = y[idx]
        # Save train and test sets of iteration i
        split = dict()
        split["x_train"] = x_train
        split["y_train"] = y_train
        split["x_test"] = x_test
        split["y_test"] = y_test
        k_fold_iter.append(split)
    return k_fold_iter




import numpy as np

def __charpath(D,diagonal_dist=None,infinite_dist=None):
    """
    Path length, global efficiency and others
    
    Note: The input distance matrix require distance_wei. Path length is the 
    mean of all the nodal path lengths. Infinite paths (disconected nodes) are
    included by default. The parameter infinite_dist may madify this

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    diagonal_dist : numpy array
        Distances on the main diagonal. Optional. Default: diagonal_dist = 0        
    infinite_dist : numpy array
        Include infinite distances in calculations. Optional. Default: infinite_dist = 1 
        
    Returns
    -------
    global_pl : int
        Path lenght.
    efficiency : int
        Network efficiency. 
    nodal_ecc : numpy array
        Network eccentricity. 
    radius : int
        Network radius. 
    diameter : int
        Network diameter. 
        
    """
    if np.any(D == np.nan):
        print('Error: distance matrix contains nan values')
    if not diagonal_dist or diagonal_dist  == None:
        np.fill_diagonal(D,np.nan) # Fill diagonal with nans        
    if not infinite_dist and not infinite_dist == None:
        D[D==np.inf] = np.nan # Ignore infinite paths 
        
    Dv = D[np.logical_not(np.isnan(D))] # Get non-nan values
        
    # Mean of the rows of D
    global_pl = np.mean(Dv)
    
    # Efficiency: Mean of the inverse entries of D
    efficiency = np.mean(1./Dv)

    # Eccentricity of each edge
    nodal_ecc = np.nanmax(D,axis=1)

    # Graph radius
    radius = np.min(nodal_ecc)
    
    # Graph diameter
    diameter = np.max(nodal_ecc)
    
    return global_pl, efficiency, nodal_ecc, radius, diameter


def __distance_wei(L):
    """
    Distance matrix (Dijkstra's algorithm'). Contains the length of the 
    shortest path between each pair of nodes.
    
    Notes: lengths between disconected nodes are set to inf. Lengths on the 
    main diagonal are set to 0

    Parameters
    ----------
    L : numpy 2D matrix
        Connection-length matrix THIS IS NOT WEIGHTS MATRIX

    Returns
    -------
    D : numpy 2D matrix
        Distance matrix.
    B : numpy 2D matrix
        Number of edges matrix. 
    """
    n = L.shape[0]       
    D = np.ones((n,n)) * np.inf # Distance matrix
    np.fill_diagonal(D,0)
    B = np.zeros((n,n)) # Number of edges matrix
    
    for u in range(0,n):
        S = np.ones((1,n),dtype=np.int32) # Distance permanence placeholder
        L1 = np.copy(L)
        V = [u]
        while 1:
            S[0,V] = 0 # Distance u>V is now permanent
            L1[:,V] = 0 # Remove in-edges
            for v in V:
                T = np.argwhere(L1[np.squeeze(v),:]) # Neighbours of shortest nodes
                d = np.min(np.concatenate((D[u,T[:,0]][:,np.newaxis], (D[u,v]+L1[v,T[:,0]])[:,np.newaxis]),axis=1),axis=1)
                wi = np.argmin(np.concatenate((D[u,T[:,0]][:,np.newaxis], (D[u,v]+L1[v,T[:,0]])[:,np.newaxis]),axis=1),axis=1)
                D[u,T[:,0]] = d # Smallest of old/new path lengths
                ind = T[wi == 1] # Indices of the lengthed paths
                B[u,ind] = B[u,v]+1 # Increment number of edges in lengthed paths
                
            if (D[u,S[0,:].astype(bool)] == 0).all(): # Empty = all nodes reached, Inf = some nodes cannot be reached
                break
            minD = np.min(D[u,S[0,:].astype(bool)])

            if minD == np.inf: # Empty = all nodes reached, Inf = some nodes cannot be reached
                break
        
            V = np.argwhere(D[u,:]==minD)
        
    return D,B


def path_length(W,diagonal_dist=None,infinite_dist=None):
    """
    Calculates the path length and other graph integration parameters
    
    Note: L must be a connection-length matrix. One way of generating it is
    the inverse of the weight matrix.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    global_pl : int
        Path lenght.
    efficiency : int
        Network efficiency. 
    nodal_ecc : numpy array
        Network eccentricity. 
    radius : int
        Network radius. 
    diameter : int
        Network diameter. 
    nodal_d : numpy 2D matrix
        Shortest distances matrix. 
        
    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
    
    W = np.array(W)
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')
        
    if not diagonal_dist == None and not (diagonal_dist == 0 or diagonal_dist == 1 
            or diagonal_dist == True or diagonal_dist == False):
        raise ValueError('Variable "diagonal_dist" must be either 0, 1, or bool')
        
    if not infinite_dist == None and not (infinite_dist == 0 or infinite_dist == 1 
            or infinite_dist == True or infinite_dist == False):
        raise ValueError('Variable "diagonal_dist" must be either 0, 1, or bool')          
    
    if not infinite_dist == None: infinite_dist = infinite_dist.astype(bool)
    if not diagonal_dist == None: diagonal_dist = diagonal_dist.astype(bool)
    
    L = np.divide(1,W)
    L[L==np.inf] = 0    
    nodal_d,_ = __distance_wei(L)    
    global_pl,efficiency,nodal_ecc,radius,diameter = __charpath(np.copy(nodal_d))
    
    return global_pl,efficiency,nodal_ecc,radius,diameter,nodal_d



import tensorflow as tf
import numpy as np
from medusa.graph_theory import degree

def __density_gpu(W):
    """
    Calculates the graph density using GPU.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    global_den : int
        Global density.   
    nodal_den : numpy array
        Nodal density. 

    """    
    
    check_symmetry = tf.reduce_all(tf.math.equal(W,tf.transpose(W)))
    check_symmetry = tf.cond(
        tf.math.reduce_sum(tf.subtract(tf.linalg.band_part(W, -1, 0),tf.linalg.band_part(W, 0, 0))) == 0,
        lambda: 1, lambda: check_symmetry)
    check_symmetry = tf.cond(
        tf.reduce_all(tf.math.equal(W,-tf.transpose(W))),
        lambda: 2, lambda: check_symmetry)
    
    N = W.shape[0]
    deg = degree.degree(W,'GPU')
    
    norm_value = tf.switch_case(tf.cast(check_symmetry,tf.int32), 
                branch_fns={0: lambda: tf.math.multiply(N,tf.math.subtract(N,1)),
                            1: lambda: tf.math.divide(tf.math.multiply(N,tf.math.subtract(N,1)),2), 
                            2: lambda: tf.math.multiply(N,tf.math.subtract(N,1))})

    nodal_den = tf.divide(deg,norm_value)
    global_den = tf.divide(tf.math.reduce_sum(deg),norm_value)     
      
    return global_den,nodal_den


def __density_cpu(W):
    """
    Calculates the graph density using CPU.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    global_den : int
        Global density.   
    nodal_den : numpy array
        Nodal density.    

    """    
    
    check_symmetry = (W.transpose() == W).all() # if symmetric
    
    if (W == np.triu(W)).all(): # if upper triangular
        check_symmetry = 1
        
    if (W.transpose() == -W).all(): # if anti-symmetric
        check_symmetry = 2
    
    N = W.shape[0]
    deg = degree.degree(W,'CPU')
    
    if check_symmetry == 0 or check_symmetry == 2:
        norm_value = N*(N-1)
    elif check_symmetry == 1:
        norm_value = ((N*(N-1))/2)
        
    nodal_den = np.divide(deg,norm_value)
    global_den = np.divide(np.sum(deg),norm_value)
      
    return global_den,nodal_den


def density(W,mode):
    """
    Calculates the graph density.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    mode : string
        GPU or CPU
        
    Returns
    -------
    global_den : int
        Global density.   
    nodal_den : numpy array
        Nodal density. 

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')        
      
    if mode == 'CPU':
        global_den,nodal_den = __density_cpu(W)
    elif mode == 'GPU':
        global_den,nodal_den = __density_gpu(W)
    else:
        raise ValueError('Unknown mode')
        
    return global_den, nodal_den



import math


def factor(n):
    """ Computes an irreducible factorization of a number.

    Parameter
    ----------
    n : int
        Number to factorize

    Returns
    ---------
    list:
        List of factors
    """
    f = lambda n: (p := [
        next(i for i in range(2, n + 1) if n % i == 0)] if n > 1 else []) + (
                      f(n // p[0]) if p else [])
    return f(n)


def optimal_subplot_row_col(n_items):
    """ Computes the optimal arrangement of items in rows and columns.

    Parameter
    ----------
    n_items : int
        Number of items.

    Returns
    ---------
    int, int:
        Optimal number of rows and columns, respectively
    """
    f = factor(n_items)
    if len(f) == 1:
        f = factor(n_items - 1)
        m_ = round(len(f) / 2)
        sub_c = math.prod(f[:m_]) + 1
        sub_r = math.prod(f[m_:])
    else:
        m_ = round(len(f) / 2)
        sub_c = math.prod(f[:m_])
        sub_r = math.prod(f[m_:])
    return sub_r, sub_c



"""
Created on Thu Aug 25 17:09:18 2022

@author: Diego Marcos-Martínez
"""
import numpy as np
import matplotlib.pyplot as plt

def time_plot(epoch, fs=1.0, ch_labels=None):
    """
    Parameters
    ---------
    epoch: numpy array
        Signal with shape of [n_epochs,n_samples, n_channels] or
        [n_samples, n_channels]
    fs: float
        Sampling rate. Value 1 as default
    ch_labels: list of strings or None
        List containing the channel labels
    """

    try:
        # Check if signals is divided in epochs
        if len(epoch.shape) == 2:
            epoch = epoch[np.newaxis,:,:]
        blocks, samples_per_block, channels = epoch.shape
        if ch_labels is None:
            ch_labels = [f"Channel {idx}" for idx in range(channels)]


        epoch_c = __reshape_signal(epoch)

        channel_offset = np.zeros(epoch_c.shape[1])
        if len(channel_offset) > 1:
            channel_offset[1:] = np.max(np.max(epoch_c[:, 1:], axis=0) -
                                        np.min(epoch_c[:, :-1], axis=0))
            channel_offset = np.cumsum(channel_offset)
        epoch_c = epoch_c - channel_offset
        max_val, min_val = epoch_c.max(), epoch_c.min()
        display_times = np.linspace(0, int(epoch_c.shape[0] / fs),
                                    epoch_c.shape[0])

        # Plot
        plt.plot(display_times, epoch_c, 'k', linewidth=0.5)
        plt.yticks(-channel_offset, labels=ch_labels)
        vertical_lines = np.empty((blocks - 1, 2, 50))
        for block in range(blocks - 1):
            vertical_lines[block, :, :] = np.asarray(
                [np.ones(50) * (block + 1) * int(samples_per_block / fs),
                 np.linspace(min_val, max_val, 50)])
            plt.plot(vertical_lines[block, 0, :], vertical_lines[block, 1, :],
                     '--', color='red', linewidth=1.5)
        plt.show(block=True)

    except Exception as e:
        print(e)

def __reshape_signal(epochs):
    """This is an auxiliary function than reshapes the signal it is divided in
    epochs in order to plot it in a row"""
    try:
        epoch_c = epochs.copy()
        blocks, samples_per_block, channels = epoch_c.shape
        epoch_c = np.reshape(epoch_c, (int(blocks * samples_per_block), channels))
        return epoch_c
    except Exception as e:
        print(e)




def request_files_to_dict(request):
    """ Extracts the files within a flask request corresponding to parameter key and and returns them in a list"""
    files = dict()
    for key in request.files.keys():
        files[key] = list()
        for value in request.files.getlist(key):
            files[key].append(value.read())

    return files



import numpy as np
# from numba import jit
import warnings
from medusa.utils import check_dimensions


def signal_orthogonalization_cpu(signal_1, signal_2):

    """ This method implements the ortogonalization of each channel of signal_1
    regarding all the channels in signal_2 using CPU

    REFERENCES: O’Neill, G. C., Barratt, E. L., Hunt, B. A., Tewarie, P. K., &
    Brookes, M.J. (2015). Measuring electrophysiological connectivity by power
    envelope correlation: a technical review on MEG methods. Physics in Medicine
    & Biology, 60(21), R271.

    Parameters
    ----------
    signal_1 : numpy 2D matrix
        First MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels],
        [n_samples, n_channels] and [n_samples].
    signal_2 : numpy 2D matrix
        Second MEEG Signal. If empty signal_2 will be set to be equal to
        signal_1. Allowed dimensions: [n_epochs, n_samples, n_channels],
        [n_samples, n_channels] and [n_samples].

    Returns
    -------
    signal_ort : numpy 3D matrix
        MEEG ortogonalised signals. The first dimension is epochs. The second
        dimension is samples. The fourth dimension is the channel of first signal,
        and the second dimension is regarding which channel of second signal it
        has been orthogonalized the channel in third dimension.
        [n_epochs, n_samples, n_channels, n_channels].

    """
    # Error check
    if signal_2 is None:
        signal_2 = signal_1
    if len(signal_1) != len(signal_2):
        raise ValueError("Signals must have the same length")
    if (type(signal_1) != np.ndarray) or (type(signal_2) != np.ndarray):
        raise ValueError("Parameter signal_1 and signal_2 must be of type "
                         "numpy.ndarray")

    # Set correct dimensions
    signal_1 = np.ascontiguousarray(check_dimensions(signal_1))
    signal_2 = np.ascontiguousarray(check_dimensions(signal_2))

    n_chan_1 = signal_1.shape[2]
    n_chan_2 = signal_2.shape[2]
    n_samp_1 = signal_1.shape[1]
    n_epo_1 = signal_1.shape[0]

    signal_1_chan = np.transpose(np.repeat(signal_1, repeats=n_chan_1, axis=2), (0,2,1))
    signal_2_chan = np.transpose(np.tile(signal_2, (1, 1, n_chan_2)),(0,2,1))

    # Add extra dimensions
    signal_1_chan = signal_1_chan[:,:,np.newaxis,:]
    signal_2_chan = signal_2_chan[:,:,np.newaxis,:]

    # Calculate pseudo inverse
    inv_signal_1_chan = np.repeat(np.linalg.pinv(np.transpose(signal_1,(0,2,1))[:, :,
                                       np.newaxis, :]),repeats=n_chan_1,axis=1)
    beta = np.matmul(signal_2_chan, inv_signal_1_chan)

    signal_ort = np.transpose(np.reshape(signal_2_chan - beta * signal_1_chan,
                            (n_epo_1,n_chan_2,n_chan_1,n_samp_1),order='F'),(0,3,1,2))

    signal_ort[:,:,np.arange(n_chan_2),np.arange(n_chan_1)] = signal_1

    return signal_ort

def signal_orthogonalization_gpu(signal_1, signal_2):
    """
    This function ortogonalizes each channel of signal_1 regarding all the
    channels in signal_2. Based in O'Neill et al. 2015

    Parameters
    ----------
    signal_1 : numpy 3D matrix
        First MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels],
        [n_samples, n_channels] and [n_samples].
    signal_2 : numpy 3D matrix
        Second MEEG Signal. If empty signal_2 will be set to be equal to
        signal_1. Allowed dimensions: [n_epochs, n_samples, n_channels],
        [n_samples, n_channels] and [n_samples].

    Returns
    -------
    signal_ort : numpy 3D matrix
        MEEG ortogonalized signals. The first dimension is epochs. The second
        dimension is samples. The fourth dimension is the channel of first signal,
        and the second dimension is regarding which channel of second signal it
        has been orthogonalized the channel in third dimension.
        [n_epochs, n_samples, n_channels, n_channels].
    """
    import tensorflow as tf

    # Error check
    if signal_2 is None:
        signal_2 = signal_1
    if len(signal_1) != len(signal_2):
        raise ValueError("Signals must have the same length")
    if (type(signal_1) != np.ndarray) or (type(signal_2) != np.ndarray):
        raise ValueError("Parameter signal_1 and signal_2 must be of type "
                         "numpy.ndarray")

    # Set correct dimensions
    signal_1 = np.ascontiguousarray(check_dimensions(signal_1))
    signal_2 = np.ascontiguousarray(check_dimensions(signal_2))

    n_chan_1 = signal_1.shape[2]
    n_chan_2 = signal_2.shape[2]
    n_samp_1 = signal_1.shape[1]
    n_epo_1 = signal_1.shape[0]

    signal_1_chan = tf.transpose(tf.repeat(signal_1, repeats=n_chan_1, axis=2),
                                 perm=[0, 2, 1])
    signal_2_chan = tf.transpose(tf.tile(signal_2,(1,1,n_chan_2)),perm=[0,2,1])

    # Add extra dimensions
    signal_1_chan = signal_1_chan[:,:,tf.newaxis,:]
    signal_2_chan = signal_2_chan[:,:,tf.newaxis,:]

    # Calculate pseudo inverse
    inv_signal_1_chan = tf.repeat(
        tf.linalg.pinv(tf.transpose(signal_1, perm=[0, 2, 1])[:, :,
                       np.newaxis, :]), repeats=n_chan_1, axis=1)
    beta = np.matmul(signal_2_chan, inv_signal_1_chan)

    signal_ort = tf.transpose(tf.transpose(tf.reshape(tf.transpose(signal_2_chan - beta *
                                         signal_1_chan,perm=[0,2,3,1]),(n_epo_1,
                                                                        n_samp_1,
                                                                        n_chan_2,
                                                                        n_chan_1)),
                              perm =(0,3,2,1)),perm=[0,3,1,2]).numpy()

    signal_ort[:,:,tf.range(n_chan_2),tf.range(n_chan_1)] = signal_1

    return signal_ort


def signal_orthogonalization_gpu_old(signal_1, signal_2):
    """
    DEPRECATED: This functions performs the orthogonalization of the signal
    but it is not vectorized nor allow epochs as inputs. Its use is not
    recommended.

    This function ortogonalizes each channel of signal_1 regarding all the
    channels in signal_2. Based in O'Neill et al. 2015

    Parameters
    ----------
    signal_1 : numpy 2D matrix
        MEEG Signal. SamplesXChannel.
    signal_2 : numpy 2D matrix
        MEEG Signal. SamplesXChannel.

    Returns
    -------
    signal_ort : numpy 3D matrix
        MEEG ortogonalised signals. Samples x Channel x Channel.

        The first dimension are the samples, the third is the base channel
        used to ortogonalize the other channels and the second dimension are
        the ortogonalized channels regarding the third dimension
    """

    import tensorflow as tf
    # Error check
    if signal_2 is None:
        signal_2 = signal_1
    if len(signal_1) != len(signal_2):
        raise ValueError("Signals must have the same length")
    if (type(signal_1) != np.ndarray) or (type(signal_2) != np.ndarray):
        raise ValueError("Parameter signal_1 and signal_2 must be of type "
                         "numpy.ndarray")

    # Variable initialization
    signal_ort = tf.zeros((len(signal_1), len(signal_2[0]), len(signal_1[0])),
                          dtype=tf.dtypes.float64)
    i1 = tf.range(0, len(signal_1))
    i2 = tf.ones_like(i1)
    i3 = tf.ones_like(i1)

    # Orthogonalization process
    signal_1_tf = tf.convert_to_tensor(signal_1)
    signal_2_tf = tf.convert_to_tensor(signal_2)

    for chan_1 in range(0, len(signal_1[0])):

        for chan_2 in range(0, len(signal_2[0])):

            signal_1_chan = tf.slice(signal_1_tf,
                                     [0, chan_1],
                                     [len(signal_1), 1])
            signal_2_chan = tf.slice(signal_2_tf,
                                     [0, chan_2],
                                     [len(signal_2), 1])
            if not np.array_equal(signal_1_chan.numpy(), signal_2_chan.numpy()):
                beta = tf.linalg.matmul(
                    tf.transpose(signal_2_chan),
                    tf.transpose(tf.linalg.pinv(signal_1_chan))
                )
                tmp = tf.math.subtract(
                    signal_2_chan, tf.math.multiply(beta,signal_1_chan)
                )
                idx = tf.stack([i1, chan_2 * i2, chan_1 * i3], axis=-1)
                signal_ort = tf.tensor_scatter_nd_update(
                    signal_ort, idx, tf.squeeze(tmp)
                )
            else:
                idx = tf.stack([i1, chan_2 * i2, chan_1 * i3], axis=-1)
                signal_ort = tf.tensor_scatter_nd_update(
                    signal_ort, idx, tf.squeeze(signal_1_chan)
                )
    return signal_ort.numpy()



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 20 15:34:18 2019
Edited on Mon Jun 13 10:00:00 2022

@author: VICTOR
@editor: Sergio Pérez-Velasco
"""
from medusa import frequency_filtering as ff
from medusa import spatial_filtering as sf
# from medusa.storage.medusa_data import MedusaData
from medusa.components import Recording
# from medusa.bci.mi_feat_extraction import extract_mi_trials_from_midata
from medusa.local_activation import statistics
# from medusa.bci.mi_models import MIModelSettings
from medusa.plots import topographic_plots
from medusa.bci.mi_paradigms import StandardPreprocessing, \
    StandardFeatureExtraction, MIDataset

import numpy as np
from scipy.ndimage import uniform_filter1d
import matplotlib.pyplot as plt
import scipy.signal as scisig


class MIPlots():
    def __init__(self, fir_order=1000, fir_cutoff=(5, 35),
                 fir_btype="bandpass", fir_method="filtfilt",
                 w_baseline_t=(-1000, 0), w_epoch_t=(-1000, 6000), norm="z",
                 baseline_mode="trial", target_fs=None):
        self.fir_order = fir_order
        self.fir_cutoff = fir_cutoff
        self.fir_btype = fir_btype
        self.fir_method = fir_method
        self.w_baseline_t = w_baseline_t
        self.w_epoch_t = w_epoch_t
        self.norm = norm
        self.target_fs = target_fs
        self.baseline_mode = baseline_mode

        # generated
        self.dataset = None
        self.fs = None
        self.channel_set = None
        self.fir = None
        self.features = None
        self.track_info = None

        self.set_sizes()

    def set_sizes(self, label_size=6, axes_size=5, line_width=1):
        self.label_size = label_size
        self.axes_size = axes_size
        self.line_width = line_width

    def extract_features(self, files):
        # Load files
        rec = Recording.load(files[0])
        self.fs = rec.eeg.fs
        self.channel_set = rec.eeg.channel_set
        self.dataset = MIDataset(channel_set=self.channel_set, fs=self.fs,
                                 experiment_att_key='midata',
                                 biosignal_att_key='eeg',
                                 experiment_mode='train')
        for file in files:
            self.dataset.add_recordings(Recording.load(file))

        # Pre-processing
        self.fir = ff.FIRFilter(order=self.fir_order, cutoff=self.fir_cutoff,
                                btype=self.fir_btype,
                                filt_method=self.fir_method)
        self.fir.fit(fs=self.fs)
        for rec in self.dataset.recordings:
            eeg = getattr(rec, self.dataset.biosignal_att_key)
            eeg.signal = self.fir.transform(signal=eeg.signal)
            eeg.signal = sf.car(signal=eeg.signal)
            setattr(rec, self.dataset.biosignal_att_key, eeg)

        # Feature extraction
        feature_extractor = StandardFeatureExtraction(
            w_epoch_t=self.w_epoch_t, target_fs=self.target_fs,
            baseline_mode=self.baseline_mode, w_baseline_t=self.w_baseline_t,
            norm=self.norm)
        self.features, self.track_info = feature_extractor.transform_dataset(
            dataset=self.dataset)

    def plot_spectrogram(self, ch_to_plot, axs_to_plot=None,
                         welch_seg_len_pct=50,
                         welch_overlap_pct=75,
                         mov_mean_hz=0):
        if self.dataset is None:
            raise Exception("Call MiPlots._extract_features() before plotting!")
        if axs_to_plot is None:
            axs_to_plot = list()
            for c in ch_to_plot:
                fig = plt.figure(figsize=(5, 5), dpi=300)
                gs = fig.add_gridspec(2, 1, wspace=0.2, hspace=0.2,
                                      height_ratios=[1, 0.2])
                axs_to_plot.append({'freq': fig.add_subplot(gs[0, 0]),
                                    'r2': fig.add_subplot(gs[1, 0])})

        labels = self.track_info["mi_labels"]
        labels_info = self.track_info["mi_labels_info"][0]

        # Compute the spectrogram
        trials_specgram = None
        new_fs = self.fs if self.target_fs is None else self.target_fs
        for t in self.features:
            welch_seg_len = np.round(
                welch_seg_len_pct / 100 * t.shape[0]).astype(int)
            welch_overlap = np.round(
                welch_overlap_pct / 100 * welch_seg_len).astype(int)
            t_freqs, t_times, t_sxx = scisig.spectrogram(
                t, fs=new_fs, axis=0, nperseg=int(256 / 2)
            )
            t_sxx = np.expand_dims(t_sxx, axis=0)
            trials_specgram = np.concatenate((trials_specgram, t_sxx), axis=0) \
                if trials_specgram is not None else t_sxx

        # Separate the classes
        trials_specgram_c1 = trials_specgram[labels == 0, :, :, :]
        trials_specgram_c2 = trials_specgram[labels == 1, :, :, :]

        # Plot ranges
        lims = [0, new_fs / 2]
        if self.fir_btype == 'bandpass':
            lims = [self.fir_cutoff[0],
                    self.fir_cutoff[1]]
        elif self.fir_btype == 'highpass':
            lims[0] = self.fir_cutoff[0]
        elif self.fir_btype == 'lowpass':
            lims[1] = self.fir_cutoff[1]

        # Plot
        lcha = self.dataset.channel_set.l_cha
        for n in range(len(ch_to_plot)):
            if ch_to_plot[n] not in lcha:
                raise ValueError('Channel ' + ch_to_plot[n] + ' is missing!')
            i = lcha.index(ch_to_plot[n])

            # Signed r2
            temp_c1 = np.squeeze(trials_specgram_c1[:, :, i, :])  # obs, t, f
            temp_c2 = np.squeeze(trials_specgram_c2[:, :, i, :])  # obs, t, f
            temp_r2 = statistics.signed_r2(temp_c1, temp_c2, signed=True,
                                           axis=0)
            with plt.style.context('seaborn'):
                # Averaged curves
                ax1 = axs_to_plot[n]
                ax1.minorticks_on()
                ax1.pcolormesh(t_times, t_freqs, temp_r2, cmap='RdBu_r')
                # ax1.grid(b=True, which='minor', color='#ededed', linestyle='--')
                # ax1.grid(b=True, which='major')
                # ax1.plot(freqs, m_psd_c1[:, i], linewidth=self.line_width,
                #          color=[255 / 255, 174 / 255, 0 / 255])
                # ax1.plot(freqs, m_psd_c2[:, i], linewidth=self.line_width,
                #          color=[24 / 255, 255 / 255, 73 / 255])
                ax1.set_ylim(lims)
                ax1.set_xlim((self.w_baseline_t[1] / 1000,
                              self.w_epoch_t[1] / 1000))
                ax1.set_title(ch_to_plot[n], fontsize=self.label_size)
                ax1.set_ylabel(r'Frequency (Hz)', fontsize=self.label_size)
                ax1.set_xlabel(r'Time (s)', fontsize=self.label_size)

                ax1.tick_params(axis='x', labelsize=self.axes_size)
                ax1.tick_params(axis='y', labelsize=self.axes_size)
        return axs_to_plot

    def plot_erd_ers_freq(self, ch_to_plot, axs_to_plot=None,
                          welch_seg_len_pct=50,
                          welch_overlap_pct=75,
                          mov_mean_hz=0):
        if self.dataset is None:
            raise Exception("Call MiPlots._extract_features() before plotting!")
        if axs_to_plot is None:
            axs_to_plot = list()
            for c in ch_to_plot:
                fig = plt.figure(figsize=(5, 5), dpi=300)
                gs = fig.add_gridspec(2, 1, wspace=0.2, hspace=0.2,
                                      height_ratios=[1, 0.2])
                axs_to_plot.append({'freq': fig.add_subplot(gs[0, 0]),
                                    'r2': fig.add_subplot(gs[1, 0])})

        labels = self.track_info["mi_labels"]
        labels_info = self.track_info["mi_labels_info"][0]

        # Compute the PSD
        trials_psd = None
        new_fs = self.fs if self.target_fs is None else self.target_fs
        for t in self.features:
            # Compute PSD of the trial
            welch_seg_len = np.round(
                welch_seg_len_pct / 100 * t.shape[0]).astype(int)
            welch_overlap = np.round(
                welch_overlap_pct / 100 * welch_seg_len).astype(int)
            t_freqs, t_psd = scisig.welch(t, fs=new_fs, nperseg=welch_seg_len,
                                          noverlap=welch_overlap,
                                          nfft=welch_seg_len, axis=0)

            # Concatenate
            t_psd = t_psd.reshape(1, t_psd.shape[0], t_psd.shape[1])
            trials_psd = np.concatenate((trials_psd, t_psd), axis=0) if \
                trials_psd is not None else t_psd

        # Separate the classes
        trials_psd_c1 = trials_psd[labels == 0, :, :]
        trials_psd_c2 = trials_psd[labels == 1, :, :]

        # Signed r2
        trials_r2 = statistics.signed_r2(trials_psd_c1, trials_psd_c2,
                                         signed=False, axis=0)
        if mov_mean_hz != 0:
            size = int(trials_psd_c1.shape[1] /
                       (self.fir_cutoff[1] - self.fir_cutoff[0]) * mov_mean_hz)
            trials_r2 = uniform_filter1d(trials_r2, size, axis=0,
                                         mode='nearest')

        # Mean PSD
        m_psd_c1 = np.mean(trials_psd_c1, axis=0)
        m_psd_c2 = np.mean(trials_psd_c2, axis=0)

        # Plot ranges
        freqs = np.linspace(0, new_fs / 2, len(m_psd_c1))
        lims = [0, new_fs / 2]
        if self.fir_btype == 'bandpass':
            lims = [self.fir_cutoff[0],
                    self.fir_cutoff[1]]
        elif self.fir_btype == 'highpass':
            lims[0] = self.fir_cutoff[0]
        elif self.fir_btype == 'lowpass':
            lims[1] = self.fir_cutoff[1]

        # Plot
        lcha = self.dataset.channel_set.l_cha
        for n in range(len(ch_to_plot)):
            if ch_to_plot[n] not in lcha:
                raise ValueError('Channel ' + ch_to_plot[n] + ' is missing!')
            i = lcha.index(ch_to_plot[n])

            with plt.style.context('seaborn'):
                # Averaged curves
                ax1 = axs_to_plot[n]['freq']
                ax1.minorticks_on()
                ax1.grid(b=True, which='minor', color='#ededed', linestyle='--')
                ax1.grid(b=True, which='major')
                ax1.plot(freqs, m_psd_c1[:, i], linewidth=self.line_width,
                         color=[255 / 255, 174 / 255, 0 / 255])
                ax1.plot(freqs, m_psd_c2[:, i], linewidth=self.line_width,
                         color=[24 / 255, 255 / 255, 73 / 255])
                ax1.set_xlim(lims)
                ax1.set_title(ch_to_plot[n], fontsize=self.label_size)
                ax1.set_ylabel(r'PSD ($uV^2/Hz$)', fontsize=self.label_size)
                ax1.legend([labels_info[str(0)], labels_info[str(1)]],
                           fontsize=self.label_size)
                ax1.tick_params(axis='x', labelsize=self.axes_size)
                ax1.tick_params(axis='y', labelsize=self.axes_size)

                # Signed-r2
                ax2 = axs_to_plot[n]['r2']
                ax2.pcolormesh(freqs, range(2),
                               np.tile(trials_r2[:, i], reps=[2, 1]),
                               cmap='YlOrRd',
                               vmin=0)
                ax2.set_xlim(lims)
                ax2.set_ylabel('$r^2$', fontsize=self.label_size)
                ax2.set_xlabel('Frequency (Hz)', fontsize=self.label_size)
                ax2.tick_params(axis='x', labelsize=self.axes_size)
                ax2.tick_params(axis='y', labelsize=self.axes_size)
        return axs_to_plot

    def plot_erd_ers_time(self):
        pass

    def plot_erd_ers_r2_topo(self, ch_to_plot, ax_to_plot=None,
                             welch_seg_len_pct=50,
                             welch_overlap_pct=75):
        if self.dataset is None:
            raise Exception("Call MiPlots._extract_features() before plotting!")
        if len(ch_to_plot) != 2:
            raise Exception("We need exactly two channels to compute r2 topo!")
        if ax_to_plot is None:
            ax_to_plot = list()
            for c in ch_to_plot:
                fig = plt.figure(figsize=(5, 5), dpi=300)
                ax_to_plot = fig.add_subplot(111)
        lcha = self.dataset.channel_set.l_cha
        labels = self.track_info["mi_labels"]
        labels_info = self.track_info["mi_labels_info"][0]

        # Compute the PSD
        trials_psd = None
        new_fs = self.fs if self.target_fs is None else self.target_fs
        for t in self.features:
            # Compute PSD of the trial
            welch_seg_len = np.round(
                welch_seg_len_pct / 100 * t.shape[0]).astype(int)
            welch_overlap = np.round(
                welch_overlap_pct / 100 * welch_seg_len).astype(int)
            t_freqs, t_psd = scisig.welch(t, fs=new_fs, nperseg=welch_seg_len,
                                          noverlap=welch_overlap,
                                          nfft=welch_seg_len, axis=0)

            # Concatenate
            t_psd = t_psd.reshape(1, t_psd.shape[0], t_psd.shape[1])
            trials_psd = np.concatenate((trials_psd, t_psd), axis=0) if \
                trials_psd is not None else t_psd

        # Separate the classes
        trials_psd_c1 = trials_psd[labels == 0, :, :]
        trials_psd_c2 = trials_psd[labels == 1, :, :]

        # Signed r2
        trials_r2 = statistics.signed_r2(trials_psd_c1, trials_psd_c2,
                                         signed=True, axis=0)
        trials_r2 = np.mean(trials_r2, axis=0)
        max_r2 = np.abs(np.max(trials_r2.flatten()))

        # Topoplot
        values = trials_r2.reshape(1, len(lcha))
        _, ax_to_plot, p_interp = topographic_plots.plot_topography(
            self.dataset.channel_set, values, clim=(-max_r2, max_r2),
            cmap='RdBu_r', linewidth=self.line_width * 2,
            head_radius=1.0, axes=ax_to_plot, show_colorbar=False,
            show=False, plot_skin_in_color=True)
        ax_to_plot.set_title("Signed $r^2$ (%s)" % ' vs. '.join(ch_to_plot),
                             fontsize=self.label_size)
        return ax_to_plot, p_interp


def _extract_erd_ers_features(files, ch_to_plot, order=1000, cutoff=[5, 35],
                              btype='bandpass', temp_filt_method='filtfilt',
                              w_epoch_t=(-1000, 6000), target_fs=None,
                              baseline_mode='trial', w_baseline_t=(-1000, 0),
                              norm='z'):
    saved_args = locals()
    del saved_args['files']
    del saved_args['ch_to_plot']

    # Load files
    rec = Recording.load(files[0])
    fs = rec.eeg.fs
    channel_set = rec.eeg.channel_set
    dataset = MIDataset(channel_set=channel_set, fs=rec.eeg.fs,
                        experiment_att_key='midataold',
                        biosignal_att_key='eeg', experiment_mode='train')
    for file in files:
        dataset.add_recordings(Recording.load(file))

    # Pre-processing
    fir = ff.FIRFilter(order=order, cutoff=cutoff, btype=btype,
                       filt_method=temp_filt_method)
    fir.fit(fs=fs)
    for rec in dataset.recordings:
        eeg = getattr(rec, dataset.biosignal_att_key)
        eeg.signal = fir.transform(signal=eeg.signal)
        eeg.signal = sf.car(signal=eeg.signal)
        setattr(rec, dataset.biosignal_att_key, eeg)

    # Feature extraction
    feature_extractor = StandardFeatureExtraction(w_epoch_t=w_epoch_t,
                                                  target_fs=target_fs,
                                                  baseline_mode=baseline_mode,
                                                  w_baseline_t=w_baseline_t,
                                                  norm=norm)
    features, track_info = feature_extractor.transform_dataset(dataset=dataset)
    lcha = dataset.channel_set.l_cha
    return features, track_info, fs, lcha, channel_set, saved_args


def plot_erd_ers_time(files, ch_to_plot, features=None, track_info=None,
                      fs=None, lcha=None, channel_set=None, mov_mean_ms=1000,
                      **kwargs):
    """Plotting function of ERD/ERS from motor imagery runs of MEDUSA.
    Parameters
    ----------
    files: list
        List of paths pointing to MI files.
    ch_to_plot: list
        List with the labels of the channels to plot
    """
    for key, value in kwargs.items():
        globals()[key] = value

    # Extract only if required
    if features is None:
        features, track_info, fs, lcha, channel_set, saved_args = \
            _extract_erd_ers_features(
                files, ch_to_plot, **kwargs
            )
        for key, value in saved_args.items():
            globals()[key] = value

    labels = track_info["mi_labels"]
    # todo: hardcoded
    labels_info = track_info["mi_labels_info"][0]
    new_fs = fs if target_fs is None else target_fs

    # # Baseline parameters
    t_baseline = [w_baseline_t[0] - w_epoch_t[0],
                  w_baseline_t[1] - w_epoch_t[0]]
    idx_baseline = np.round(np.array(t_baseline) * new_fs / 1000).astype(int)

    # Separate the classes
    trials_c1 = features[labels == 0, :, :]
    trials_c2 = features[labels == 1, :, :]

    # Compute the average power
    p_c1 = np.power(trials_c1, 2)
    p_c2 = np.power(trials_c2, 2)
    p_c1_avg = np.mean(p_c1, axis=0)
    p_c2_avg = np.mean(p_c2, axis=0)

    # Compute the reference power for each channel
    r_c1_mean = np.mean(p_c1_avg[idx_baseline[0]:idx_baseline[1], :], axis=0)
    r_c2_mean = np.mean(p_c2_avg[idx_baseline[0]:idx_baseline[1], :], axis=0)

    # Compute ERD/ERS
    ERDERS_c1 = 100 * (p_c1_avg - r_c1_mean) / r_c1_mean
    ERDERS_c2 = 100 * (p_c2_avg - r_c2_mean) / r_c2_mean
    # TODO: Cambiar por https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.uniform_filter1d.html

    ERDERS_c1_smooth = uniform_filter1d(ERDERS_c1,
                                        int(np.floor(
                                            mov_mean_ms * new_fs / 1000)),
                                        axis=0, mode='mirror')
    ERDERS_c2_smooth = uniform_filter1d(ERDERS_c2,
                                        int(np.floor(
                                            mov_mean_ms * new_fs / 1000)),
                                        axis=0, mode='mirror')

    # Signed r2 for the power
    p_c1_marg = 100 * (p_c1 - r_c1_mean / p_c1.shape[0]) / r_c1_mean
    p_c2_marg = 100 * (p_c2 - r_c2_mean / p_c2.shape[0]) / r_c2_mean
    trials_r2 = statistics.signed_r2(p_c1_marg, p_c2_marg, signed=False, axis=0)
    if mov_mean_ms != 0:
        trials_r2 = uniform_filter1d(trials_r2,
                                     int(np.floor(mov_mean_ms * new_fs / 1000)),
                                     axis=0, mode='mirror')

    # Plotting
    times = np.linspace(w_epoch_t[0], w_epoch_t[1], ERDERS_c1_smooth.shape[0])

    # Plot
    left = 0.1
    bottom = 0
    width = 0.8
    height_psd = 0.6
    height_r2 = 0.06
    height_cbar = 0.06
    gap = 0.12

    figs = list()
    for n in range(len(ch_to_plot)):
        fig = plt.figure()

        ax1 = fig.add_axes(
            [left, bottom + height_r2 + height_cbar + gap, width, height_psd],
            xticklabels=[])
        ax2 = fig.add_axes([left, bottom + height_cbar + gap, width, height_r2],
                           yticklabels=[])

        if ch_to_plot[n] not in ch_to_plot:
            raise ValueError('Channel ' + ch_to_plot[n] + ' is missing!')
        i = lcha.index(ch_to_plot[n])

        # ERD/ERS(%)
        ax1.minorticks_on()
        ax1.grid(visible=True, which='minor', color='#ededed', linestyle='--')
        ax1.grid(visible=True, which='major')
        ax1.axvline(x=0, color='k', linestyle='--', label='_nolegend_')
        ax1.axvspan(w_baseline_t[0], w_baseline_t[1], alpha=0.1,
                    facecolor='gray', label='_nolegend_')
        ax1.plot(times, ERDERS_c1_smooth[:, i], linewidth=2,
                 color=[255 / 255, 174 / 255, 0 / 255])
        ax1.plot(times, ERDERS_c2_smooth[:, i], linewidth=2,
                 color=[24 / 255, 255 / 255, 73 / 255])
        ax1.set_xlim(w_epoch_t)
        ax1.title.set_text(ch_to_plot[n])
        ax1.set_ylabel(r'ERD/ERS (%)')
        ax1.legend([labels_info[str(0)], labels_info[str(1)]])

        # Signed-r2
        ax2.pcolormesh(times, range(2), np.tile(trials_r2[:, i], reps=[2, 1]),
                       cmap='YlOrRd')
        ax2.axvline(x=0, color='k', linestyle='--', label='_nolegend_')
        ax2.set_ylabel('$r^2$')
        ax2.set_xlabel('Time (ms)')
        figs.append(fig)

    return figs


def plot_erd_ers_freq(files, ch_to_plot, features=None, track_info=None,
                      fs=None, lcha=None, channel_set=None,
                      welch_seg_len_pct=50,
                      welch_overlap_pct=75, mov_mean_hz=0,
                      **kwargs):
    # TODO: More options!...
    # TODO: Left and right classes labels are hardcoded!
    """ This function depicts the ERD/ERS events of MI BCIs over the frequency
    spectrum.
    Parameters
    ----------
    files: list
        List of paths pointing to MI files.
    ch_to_plot: list
        List with the labels of the channels to plot
    """
    for key, value in kwargs.items():
        globals()[key] = value

    # Extract only if required
    if features is None:
        features, track_info, fs, lcha, channel_set, saved_args = \
            _extract_erd_ers_features(
                files, ch_to_plot, **kwargs
            )
        for key, value in saved_args.items():
            globals()[key] = value

    labels = track_info["mi_labels"]
    # todo: hardcoded
    labels_info = track_info["mi_labels_info"][0]

    # Compute the PSD
    trials_psd = None
    new_fs = fs if target_fs is None else target_fs
    for t in features:
        # Compute PSD of the trial
        welch_seg_len = np.round(welch_seg_len_pct / 100 * t.shape[0]).astype(
            int)
        welch_overlap = np.round(
            welch_overlap_pct / 100 * welch_seg_len).astype(int)
        welch_ndft = welch_seg_len
        t_freqs, t_psd = scisig.welch(t, fs=new_fs, nperseg=welch_seg_len,
                                      noverlap=welch_overlap,
                                      nfft=welch_ndft, axis=0)

        # Concatenate
        t_psd = t_psd.reshape(1, t_psd.shape[0], t_psd.shape[1])
        trials_psd = np.concatenate((trials_psd, t_psd),
                                    axis=0) if trials_psd is not None else t_psd

    # Separate the classes
    trials_psd_c1 = trials_psd[labels == 0, :, :]
    trials_psd_c2 = trials_psd[labels == 1, :, :]

    # Signed r2
    trials_r2 = statistics.signed_r2(trials_psd_c1, trials_psd_c2, signed=False,
                                     axis=0)
    if mov_mean_hz != 0:
        size = int(
            trials_psd_c1.shape[1] / (cutoff[1] - cutoff[0]) * mov_mean_hz)
        trials_r2 = uniform_filter1d(trials_r2, size,
                                     axis=0, mode='nearest')

    # Mean PSD
    m_psd_c1 = np.mean(trials_psd_c1, axis=0)
    m_psd_c2 = np.mean(trials_psd_c2, axis=0)

    # Plot ranges
    freqs = np.linspace(0, new_fs / 2, len(m_psd_c1))
    lims = [0, new_fs / 2]
    if btype == 'bandpass':
        lims = [cutoff[0],
                cutoff[1]]
    elif btype == 'highpass':
        lims[0] = cutoff[0]
    elif btype == 'lowpass':
        lims[1] = cutoff[1]

    # Plot
    left = 0.1
    bottom = 0
    width = 0.8
    height_psd = 0.6
    height_r2 = 0.06
    height_cbar = 0.06
    gap = 0.12

    figs = list()
    for n in range(len(ch_to_plot)):
        fig = plt.figure()

        ax1 = fig.add_axes(
            [left, bottom + height_r2 + height_cbar + gap, width, height_psd],
            xticklabels=[])
        ax2 = fig.add_axes([left, bottom + height_cbar + gap, width, height_r2],
                           yticklabels=[])

        if ch_to_plot[n] not in ch_to_plot:
            raise ValueError('Channel ' + ch_to_plot[n] + ' is missing!')
        i = lcha.index(ch_to_plot[n])

        # Individual curves
        # for j in range(trials_psd_c1.shape[0]):
        #    plt.plot(freqs, trials_psd_c1[j,:,i], linewidth=0.5,
        #             color=[255/255, 174/255, 0/255], alpha=0.5)
        # for j in range(trials_psd_c2.shape[0]):
        #    plt.plot(freqs, trials_psd_c2[j,:,i], linewidth=0.5,
        #             color=[24/255, 255/255, 73/255], alpha=0.5)

        # Averaged curves
        ax1.minorticks_on()
        ax1.grid(b=True, which='minor', color='#ededed', linestyle='--')
        ax1.grid(b=True, which='major')
        ax1.plot(freqs, m_psd_c1[:, i], linewidth=2,
                 color=[255 / 255, 174 / 255, 0 / 255])
        ax1.plot(freqs, m_psd_c2[:, i], linewidth=2,
                 color=[24 / 255, 255 / 255, 73 / 255])
        ax1.set_xlim(lims)
        ax1.title.set_text(ch_to_plot[n])
        ax1.set_ylabel(r'PSD ($uV^2/Hz$)')
        ax1.legend([labels_info[str(0)], labels_info[str(1)]])

        # Signed-r2
        ax2.pcolormesh(freqs, range(2), np.tile(trials_r2[:, i], reps=[2, 1]),
                       cmap='YlOrRd', vmin=0)
        ax2.set_xlim(lims)
        ax2.set_ylabel('$r^2$')
        ax2.set_xlabel('Frequency (Hz)')
        figs.append(fig)
    return figs


def plot_r2_topoplot(files, ch_to_plot, features=None, track_info=None,
                     fs=None, lcha=None, channel_set=None,
                     welch_seg_len_pct=50,
                     welch_overlap_pct=75, background=False, **kwargs):
    for key, value in kwargs.items():
        globals()[key] = value

    # Extract only if required
    if features is None:
        features, track_info, fs, lcha, channel_set, saved_args = \
            _extract_erd_ers_features(
                files, ch_to_plot, **kwargs
            )
        for key, value in saved_args.items():
            globals()[key] = value

    labels = track_info["mi_labels"]
    # todo: hardcoded
    labels_info = track_info["mi_labels_info"][0]
    new_fs = fs if target_fs is None else target_fs

    # Compute the PSD
    trials_psd = None
    for t in features:
        # Compute PSD of the trial
        welch_seg_len = np.round(welch_seg_len_pct / 100 * t.shape[0]).astype(
            int)
        welch_overlap = np.round(
            welch_overlap_pct / 100 * welch_seg_len).astype(int)
        welch_ndft = welch_seg_len
        t_freqs, t_psd = scisig.welch(t, fs=new_fs, nperseg=welch_seg_len,
                                      noverlap=welch_overlap,
                                      nfft=welch_ndft, axis=0)

        # Concatenate
        t_psd = t_psd.reshape(1, t_psd.shape[0], t_psd.shape[1])
        trials_psd = np.concatenate((trials_psd, t_psd),
                                    axis=0) if trials_psd is not None else t_psd

    # Separate the classes
    trials_psd_c1 = trials_psd[labels == 0, :, :]
    trials_psd_c2 = trials_psd[labels == 1, :, :]

    # Signed r2
    trials_r2 = statistics.signed_r2(trials_psd_c1, trials_psd_c2, signed=True,
                                     axis=0)
    trials_r2 = np.mean(trials_r2, axis=0)

    # Topoplot
    values = trials_r2.reshape(1, len(lcha))
    fig, _, _ = topographic_plots.plot_topography(channel_set,
                                                  values, cmap='RdBu',
                                                  background=background,
                                                  show=False)

    return fig



from medusa import components
from medusa import meeg


def test_load_eeg_recording():
    rec = components.Recording.load('data/eeg.rec.json')
    assert isinstance(rec, components.Recording)
    return rec


if __name__ == '__main__':
    rec = test_load_eeg_recording()




from setuptools import setup, find_packages
from pathlib import Path

# Just testing

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='medusa-kernel',
    packages=find_packages(),
    version='1.1.1',
    keywords=['Signal', 'Biosignal', 'EEG', 'BCI'],
    url='https://medusabci.com/',
    author='Eduardo Santamaría-Vázquez, '
           'Víctor Martínez-Cagigal, '
           'Diego Marcos-Martínez, '
           'Víctor Rodríguez-González, '
           'Sergio Pérez-Velasco',
    author_email='support@medusabci.com',
    license='CC Attribution-NonCommercial-NoDerivs 2.0',
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'sklearn',
        'statsmodels',
        'bson',
        'h5py',
        'dill',
        'tqdm',
        'tensorflow'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
    ],
    description='Advanced biosignal processing toolbox',
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data={
        'medusa': ['meeg/*.tsv', 'local_activation/*.dll']
    },
    license_files=('LICENSE',),
)



""" Last edit: Víctor Martínez-Cagigal (07 Dec 2021) """

import numpy as np


def itr(accuracy, n_commands, selections_per_min):
    """
    Calculates the information transfer rate (ITR) in bits/min. This metric is
    widely used to assess the performance of a BCI. However, it has serious
    limitations for online systems: (i) it assumes that the probability of
    selecting all symbols is the same; (ii) the system is memoryless; (iii) a
    synchronous paradigm is used; (iv) users cannot correct mistakes.

    Parameters
    ---------------
    accuracy: float or list of floats
        Accuracy of the system (between 0 and 1).
    n_commands: int
        Number of possible commands to be selected.
    selections_per_min: float
        Number of selections that have been performed in a minute.

    Returns
    ----------------
    itr: float or list of floats
        ITR corresponding to each accuracy in bits per min (bpm).
    """
    # Special cases
    if accuracy == 0:
        # If accuracy is 0%, then ITR is 0 bpm
        itr = 0
    elif accuracy == 1:
        # If accuracy is 100%, then we take the mathematical limit
        itr = np.log2(n_commands) * selections_per_min
    else:
        # Otherwise: common ITR formula
        itr = (np.log2(n_commands) + accuracy * np.log2(accuracy) +
               (1 - accuracy) * np.log2((1 - accuracy) / (n_commands - 1))) \
              * selections_per_min

    return itr


# Built-in imports
import math
import ctypes

# External imports
import numpy as np
from scipy.spatial.distance import pdist
from scipy.signal import decimate

# Medusa imports
from medusa.components import ThreadWithReturnValue
from medusa.utils import check_dimensions


def central_tendency_measure(signal, r):
    """
    This method implements the central tendency measure (CTM). This parameter is
    useful to quantify the variability of a signal. It is based on calculating
    the second-order differences diagram of the time series and then counting
    the points within a radius "r". CTM assigns higher values to less variable
    signals

    References
    ----------
    Cohen, M. E., Hudson, D. L., & Deedwania, P. C. (1996). Applying
    continuous chaotic modeling to cardiac signal analysis. IEEE Engineering in
    Medicine and Biology Magazine, 15(5), 97-102.

    Parameters
    ----------
    signal : numpy.ndarray
        Signal with shape [n_epochs, n_samples, n_channels].
    r : double
        Radius used to compute the CTM.

    Returns
    -------
    ctm : numpy.ndarray
        CTM values for each channel in "signal". [n_epochs, n_channels].
    """

    #  Error check
    if not np.issubdtype(signal.dtype, np.number):
        raise ValueError('data matrix contains non-numeric values')

    # Check dimensions
    signal = check_dimensions(signal)

    # Signal dimensions
    n_epo = signal.shape[0]
    n_samp = signal.shape[1]
    n_cha = signal.shape[2]

    # Values within a range (mean +- 3 std)
    upper_bound = np.mean(signal, axis=1) + 3 * np.std(signal, axis=1)
    lower_bound = np.mean(signal, axis=1) - 3 * np.std(signal, axis=1)
    idx_within_range = np.logical_and((signal < upper_bound[:, None, :]),
                                      (signal > lower_bound[:, None, :]))
    idx_out_upper = (signal > upper_bound[:, None, :])
    idx_out_lower = (signal < lower_bound[:, None, :])

    # Maximum value in the above defined range
    max_value = np.empty((n_epo, n_cha))
    for ep in range(n_epo):
        for ch in range(n_cha):
            max_value[ep, ch] = np.max(
                abs(signal[ep, idx_within_range[ep, :, ch], ch]), axis=0)

    # Normalize the values within the range by its maximum.Values above that
    # range will be 1, and below the range will be - 1
    data_norm = np.zeros_like(signal)
    data_norm[idx_within_range] = np.divide(
        signal[idx_within_range],
        np.tile(max_value, (1, n_samp, 1)).flatten()[
            idx_within_range.flatten()])
    data_norm[idx_out_upper] = 1
    data_norm[idx_out_lower] = -1

    # Difference time series
    y = data_norm[:, 3:n_samp, :] - data_norm[:, 2:n_samp - 1, :]
    x = data_norm[:, 2:n_samp - 1, :] - data_norm[:, 1:n_samp - 2, :]

    # CTM - Values below the radius 'r'
    ctm = np.sum(np.sqrt(np.square(x) + np.square(y)) < r, axis=1) / (
                n_samp - 2)

    return ctm


def sample_entropy(signal, m, r, dist_type='chebyshev'):
    """ This method implements the sample entropy (SampEn). SampEn is an
    irregularity measure that assigns higher values to more irregular time
    sequences. It has two tuning parameters: the sequence length (m) and the
    tolerance (r)

    Notes: IF A = 0 or B = 0, SamEn would return an infinite value.
    However, the lowest non-zero conditional probability that SampEn should
    report is A/B = 2/[(N-m-1)*(N-m)].
    SampEn has the following limits:
                 - Lower bound: 0
                 - Upper bound : log(N-m) + log(N-m-1) - log(2)

    References
    ----------
    Richman, J. S., & Moorman, J. R. (2000). Physiological
    time-series analysis using approximate entropy and sample entropy. American
    Journal of Physiology-Heart and Circulatory Physiology.

    Parameters
    ----------
    signal : numpy.ndarray
        Signal with shape [n_epochs, n_samples, n_channels].
    m : int
        Sequence length
    r : float
        Tolerance
    dist_type : string
        Distance metric

    Returns
    -------
    sampen : numpy.ndarray
        SampEn value. [n_epochs, n_channels]
    """
    # Check dimensions
    signal = check_dimensions(signal)

    # Check Errors
    if m > signal.shape[1]:
        raise ValueError('Embedding dimension must be smaller than the signal '
                         'length (m<N).')
    if not isinstance(dist_type, str):
        raise ValueError('Distance type must be a string.')
    if dist_type not in ['braycurtis', 'canberra', 'chebyshev', 'cityblock',
                         'correlation', 'cosine', 'dice', 'euclidean',
                         'hamming', 'jaccard', 'jensenshannon', 'kulsinski',
                         'mahalanobis', 'matching', 'minkowski',
                         'rogerstanimoto', 'russellrao', 'seuclidean',
                         'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule']:
        raise ValueError(
            'Distance type unknown. Please, check allowed distances'
            'in pdist function from scipy.spatial.distance module.')

    # Useful parameters
    n_epo = signal.shape[0]
    N = signal.shape[1]
    n_channels = signal.shape[2]
    sigma = np.std(signal, axis=1)
    templates_m = []
    templates_m_plus_one = []
    B, A, value = np.empty((n_epo, n_channels)), np.empty((n_epo, n_channels)),\
                  np.empty((n_epo, n_channels))

    # Calculate B values
    for i in range(N - m + 1):
        templates_m.append(signal[:, i:i + m, :])
    templates_m = np.array(templates_m)
    for e_idx in range(n_epo):
        w_threads = []
        for ch_idx in range(n_channels):
            t = ThreadWithReturnValue(
                target=pdist,
                args=(templates_m[:, e_idx, :, ch_idx], dist_type,))
            w_threads.append(t)
            t.start()
        for th_idx, thread in enumerate(w_threads):
            B[e_idx, th_idx] = np.sum(thread.join() <= sigma[0, th_idx] * r)

    # Check if there is any B = 0
    zeros_idx = np.where(B == 0)
    value[zeros_idx] = math.inf

    # Calculate A values
    m += 1
    for i in range(N - m + 1):
        templates_m_plus_one.append(signal[:, i:i + m, :])
    templates_m_plus_one = np.array(templates_m_plus_one)
    for e_idx in range(n_epo):
        w_threads = []
        for ch_idx in range(n_channels):
            t = ThreadWithReturnValue(
                target=pdist, args=(templates_m_plus_one[:, e_idx, :, ch_idx],
                                    dist_type,))
            w_threads.append(t)
            t.start()
        for th_idx, thread in enumerate(w_threads):
            A[e_idx, th_idx] = np.sum(thread.join() <= sigma[0, th_idx] * r)

    # Check if there is any A = 0
    zeros_idx = np.where(A == 0)
    value[zeros_idx] = math.inf

    non_inf = np.where(value != math.inf)
    value[non_inf] = -np.log((A[non_inf] / B[non_inf])*((N - m + 1) / (N - m - 1)))

    # If there is infinity values
    inf_indx = np.where(value == math.inf)
    value[inf_indx] = -np.log(2 / ((N - m - 1) * (N - m)))
    return value


def multiscale_entropy(signal, max_scale, m, r):
    """ This method implements the Multiscale Entropy (MSE). MSE is a
    measurement of complexity which measures the irregularity of a signal over
    multiple time-scales. This is accomplished through estimation of the Sample
    Entropy (SampEn) on coarse-grained versions of the original signal. As a
    result of the these calculations, MSE curves are obtained and can be used to
    compare the complexity of time-series. The MSE curve whose entropy values
    are higher for the most of time-scales is considered more complex.

    References
    ----------
    Costa, M., Goldberger, A. L., & Peng, C. K. (2005). Multiscale entropy
    analysis of biological signals, Physical review E, 71(2), 021906.

    Parameters
    ----------
    signal : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].
    max_scale : int
        Maximum scale value
    m : int
        Sequence length
    r : float
        Tolerance

    Returns
    -------
    mse_result : numpy.ndarray
        SampEn values for each scale for each channel in "signal" .
        [n_epochs, max_scale, n_channels].

    """
    # Check dimensions
    signal = check_dimensions(signal)

    # Signal dimensions
    n_epo = signal.shape[0]
    n_channels = signal.shape[2]

    mse_result = np.empty((n_epo, max_scale, n_channels))
    w_threads = list()
    scales = list()
    for i in range(1, max_scale + 1):
        if i == 1:
            t = ThreadWithReturnValue(
                target=sample_entropy,
                args=(signal, m, r, 'chebyshev'))
        else:
            t = ThreadWithReturnValue(
                target=sample_entropy,
                args=(__coarse_grain(signal, i),
                      m, r, 'chebyshev'))
        w_threads.append(t)
        scales.append(i)
        t.start()
    for t_idx, thread in enumerate(w_threads):
        mse_result[:, t_idx, :] = thread.join()
    return mse_result


def __coarse_grain(signal, scale, decimate_mode=True):
    if decimate_mode:
        return decimate(signal, scale, axis=1)
    else:
        # Signal dimensions
        n_epo = signal.shape[0]
        N = signal.shape[1]
        n_cha = signal.shape[2]

        # Number of coarse grains in which the signal is split
        tau = int(round(N / scale))

        # Returned signal
        y = np.empty((n_epo,tau,n_cha))
        for i in range(tau):
            y[:, i, :] = np.mean(signal[:, i * scale:(i * scale + scale), :])
        return y


def __lempelziv_complexity(signal):
    """ Calculate the  signal binarisation and the Lempel-Ziv's complexity. This
    version allows the algorithm to be used for signals with more than one
    channel at the same time.

    Warnings: This function is deprecated and will be removed in future updates
    of MEDUSA Kernel.

    Parameters
    ---------
    signal: numpy 2D array
        Signal with shape [n_samples, n_channels]

    Returns
    -------
    lz_channel_values: numpy 1D matrix
        Lempel-Ziv values for each channel with shape [n_channels].
    """

    if signal.size == len(signal):
        signal = signal[:, np.newaxis]

    signal = __binarisation(signal, [signal.shape[0]], signal.shape[0])
    if signal.shape[1] == 1:
        return __lz_algorithm(signal)
    else:
        lz_channel_values = np.empty((signal.shape[1]))
        working_threads = list()
        for ch in range(signal.shape[1]):
            t = ThreadWithReturnValue(target=__lz_algorithm,
                                      args=(signal[:, ch],))
            working_threads.append(t)
            t.start()
        for index, thread in enumerate(working_threads):
            lz_channel_values[index] = thread.join()
        return lz_channel_values


def lempelziv_complexity(signal):
    """Calculate the  signal binarisation and the Lempel-Ziv's complexity.
    This function takes advantage of its implementation in C to achieve a high
    performance.

   Parameters
   ---------
   signal: numpy.ndarray
       MEEG Signal [n_epochs, n_samples, n_channels]

   Returns
   -------
   lz_result: numpy.ndarray
       Lempel-Ziv values for each epoch and each channel. [n_epochs, n_channels].
    """
    # Check dimensions
    signal = check_dimensions(signal)

    # Signal dimensions
    n_epo = signal.shape[0]
    n_sample = signal.shape[1]
    n_cha = signal.shape[2]

    # Initialize result matrix
    lz_result = np.empty((n_epo, n_cha))

    # Adapt the signal to the format required by the DLL
    signal = np.reshape(np.transpose(signal, (0, 2, 1)), (n_epo, -1))

    # Get function from dll
    dll_file = ".\computeLZC.dll"
    lib = ctypes.cdll.LoadLibrary(dll_file)
    lzc_func = lib.computeLempelZivCmplx  # Access function

    for e_idx, epochs in enumerate(signal):
        # Create empty output vector
        lz_channel_values = np.zeros(int(n_cha), dtype=np.double)

        # Define inputs and outputs
        lzc_func.restype = None
        array_1d_double = np.ctypeslib.ndpointer(dtype=np.double, ndim=1,
                                                 flags='CONTIGUOUS')
        lzc_func.argtypes = [array_1d_double, array_1d_double, ctypes.c_int32,
                             ctypes.c_int32]

        # Call the function in the dll
        lzc_func(signal[e_idx, :], lz_channel_values, int(n_sample * n_cha),
                 int(n_sample))
        lz_result[e_idx, :] = lz_channel_values
    return lz_result


def multiscale_lempelziv_complexity(signal, W):
    """Calculate the multiscale signal binarisation and the Multiscale
    Lempel-Ziv's complexity.

    References
    ----------
    Ibáñez-Molina, A. J., Iglesias-Parro, S., Soriano, M. F., & Aznarte, J. I,
    Multiscale Lempel-Ziv complexity for EEG measures. Clinical Neurophysiology,
    (2015), 126(3), 541–548.

    Parameters
    ---------
    signal: numpy.ndarray
        MEEG Signal [n_epochs, n_samples, n_channels]
    W: list or numpy 1D array
        Set of window length to consider at multiscale binarisation stage.
        Values must be odd.

    Returns
    -------
    result: numpy.ndarray
        Matrix of results with shape [n_epochs, n_window_length, n_channels]

    """
    # Check dimensions
    signal = check_dimensions(signal)

    # Signal dimensions
    n_epo = signal.shape[0]
    n_cha = signal.shape[2]

    # Useful parameter
    w_max = W[-1] + (W[1] - W[0])

    # Define a matrix to store results
    result = np.empty((n_epo, len(W), n_cha))

    # First get binarised signal
    for ep_idx, epoch in enumerate(signal):
        for w_idx, w in enumerate(W):
            binarised_signal = __binarisation(epoch, w, w_max, multiscale=True)

            # Parallelize the calculations if n_channel > 1
            if binarised_signal.shape[1] > 1:
                threads = []
                for ch in range(binarised_signal.shape[1]):
                    t = ThreadWithReturnValue(target=__lz_algorithm,
                                              args=(binarised_signal[:, ch],))
                    threads.append(t)
                    t.start()
                for ch_idx, t in enumerate(threads):
                    result[ep_idx, w_idx, ch_idx] = t.join()
            else:
                result[ep_idx, w_idx, :] = __lz_algorithm(binarised_signal)
    return result


def __lz_algorithm(signal):
    """
    Lempel-Ziv's complexity algorithm implemented in Python.

    References
    ----------
    F. Kaspar, H. G. Schuster, "Easily-calculable measure for the complexity of
    spatiotemporal patterns", Physical Review A, Volume 36, Number 2 (1987).

    Parameters
    ---------
    signal: numpy 1D array
        Signal with shape of [n_samples]

    Returns
    -------
    value: float
        Result of algorithm calculations.

    """
    signal = signal.flatten().tolist()
    i, k, l = 0, 1, 1
    c, k_max = 1, 1
    n = len(signal)
    while True:
        if signal[i + k - 1] == signal[l + k - 1]:
            k = k + 1
            if l + k > n:
                c = c + 1
                break
        else:
            if k > k_max:
                k_max = k
            i = i + 1
            if i == l:
                c = c + 1
                l = l + k_max
                if l + 1 > n:
                    break
                else:
                    i = 0
                    k = 1
                    k_max = 1
            else:
                k = 1

    value = c * (np.log2(n) / n)
    return value


def __binarisation(signal, w_length, w_max, multiscale=False):
    """
    This function returns a binarised version of original signal. It can be used
    in both multiscale and simple binarisations. If multiscale mode is chosen,
    signals are shortened so that they all have the same length, taking the
    maximum window as a reference. Binarisation is performed by means of a
    median-based comparison.

    Parameters
    ----------
    signal: numpy 2D array
        Signal with shape [n_samples x n_channels]
    w_length: int
        Window length to perform multiscale binarisation
    w_max: int
        Value of the maximum window length
    multiscale: bool
        If is True, performs the multiscale binarisation

    Returns
    -------
    signal_binarised: numpy.ndarray
        Signal binarised with shape of [n_samples_shortened x n_channels]
        The n_samples_shortened parameter is calculated from w_max to ensure that
        all signals have the same length, regardless of the window length used.
    """

    if multiscale:
        if w_length is None:
            raise ValueError('Width of window must be an integer value')
        if w_length % 2 == 0:
            raise ValueError('Width of window must be an odd value.')
        if w_max is None:
            raise ValueError('Maximum window width must be an integer value')

        #  Get smoothed version from original signal
        smoothed = __multiscale_median_threshold(signal, w_length)

        # Useful parameters
        half_wind = int((w_length - 1) / 2)
        max_length = signal.shape[0] + 1 - w_max
        length_diff = smoothed.shape[0] - max_length

        # Shorten original and smoothed version
        smoothed_shortened = \
            smoothed[int(length_diff / 2):-int(length_diff / 2), :]
        signal_shortened = \
            signal[half_wind: signal.shape[0] - half_wind, :]
        signal_shortened = \
            signal_shortened[int(length_diff / 2):-int(length_diff / 2), :]

        # Define template of binarised signal
        signal_binarised = \
            np.zeros((signal_shortened.shape[0], signal_shortened.shape[1]))

        # Binarise the signal
        idx_one = signal_shortened >= smoothed_shortened
        signal_binarised[idx_one] = 1

    else:
        signal_binarised = np.zeros((len(signal), signal.shape[1]))
        median = np.median(signal, axis=0)
        idx_one = signal >= median
        signal_binarised[np.squeeze(idx_one)] = 1

    return signal_binarised


def __multiscale_median_threshold(signal, w_length):
    """ Signal smoothing function. For each sample, we define a window in which
    the sample is in centre position. The median value of the window is
    calculated and assigned to this sample to obtain a smoothed version of the
    original signal.

    References
    ----------
    Ibáñez-Molina, A. J., Iglesias-Parro, S., Soriano, M. F., & Aznarte, J. I,
    Multiscale Lempel-Ziv complexity for EEG measures. Clinical Neurophysiology,
    (2015), 126(3), 541–548.

    Parameters
    ---------
    signal: numpy 2D array
        Signal with shape of [n_samples, n_channels]
    w_length: int
        Length of window to calculate median values in smoothing process

    Returns
    -------
    smoothed_signal: numpy 2D array
        Smoothed version with shape of [n_samples + 1 - w_length, n_channels]
        As a result of windowing, the final samples of the signal are lost.

    """
    # Template of smoothed signal
    smoothed_signal = np.zeros((
        signal.shape[0] + 1 - w_length, signal.shape[1]))

    half_wind = int((w_length - 1) / 2)

    # Index of sample to be smoothed from median window value
    index = 0

    # We define a window with samp in central position and
    # get median value to smooth original signal
    for samp in range(half_wind, signal.shape[0] - half_wind):
        smoothed_signal[index, :] = np.median(
            signal[samp - half_wind: samp + half_wind], axis=0)
        index += 1
    return smoothed_signal

